"""Reverse path settings"""
from argparse import ArgumentDefaultsHelpFormatter
import subprocess, os
from . import util


def _read_filter(ifname, stdout=False):
    safe_ifname = ifname.replace(".", "/")
    cmd = ["sysctl", "net.ipv4.conf." + safe_ifname + ".rp_filter"]
    if (stdout):
        print(" ".join(cmd))
        return
    return subprocess.check_output(cmd).split()[-1].decode()


def _write_filter(ifname, val):
    assert(val == "1" or val == "0")
    safe_ifname = ifname.replace(".", "/")
    cmd = ["sysctl", "-w", "net.ipv4.conf." + safe_ifname + ".rp_filter=" + val]

    # Run if root, print if normal user
    if (os.geteuid() == 0):
        action = "Enabling" if val == "1" else "Disabling"
        print("{} RP filter on interface {}".format(action, ifname))
        subprocess.check_output(cmd)
    else:
        print(" ".join(cmd))

def _rm_filter(ifname):
    _write_filter(ifname, "0")


def _add_filter(ifname):
    _write_filter(ifname, "1")


def _check_rp_filters(dvb_if):
    """Check if reverse-path (RP) filters are configured on the interface

    Args:
        dvb_if : DVB network interface

    Return:
        True when configuration is already OK.

    """

    # Check current configuration of DVB interface and "all" rule:
    dvb_cfg = _read_filter(dvb_if)
    all_cfg = _read_filter("all")

    return (dvb_cfg == "0" and all_cfg == "0")


def _set_filters(dvb_ifs, non_root):
    """Disable reverse-path (RP) filtering for the DVB interface

    There are two layers of RP filters, one specific to the network interface
    and a higher level that controls the configurations for all network
    interfaces. This function disables RP filtering on the top layer (for all
    interfaces), but then enables RP filtering individually for each interface,
    except the DVB interface. This way, in the end only the DVB interface has RP
    filtering disabled.

    Args:
        dvb_ifs : DVB network interfaces

    """
    assert(isinstance(dvb_ifs, list))

    # Check "all" rule:
    all_cfg = _read_filter("all")

    # If "all" rule is already disabled, it is only necessary to disable the
    # target interface
    if (all_cfg == "0"):
        if (not non_root):
            print("RP filter for \"all\" interfaces is already disabled")
        for dvb_if in dvb_ifs:
            _rm_filter(dvb_if)
        if (non_root):
            print()
            util.fill_print(
                "NOTE: this assumes the RP filter for \"all\" \
                interfaces is already disabled. You can check this by running:"
            )
            _read_filter("all", stdout=True)

    # If "all" rule is enabled, we will need to disable it. Also to preserve
    # RP filtering on all other interfaces, we will enable them manually.
    else:
        # Check interfaces
        ifs = os.listdir("/proc/sys/net/ipv4/conf/")

        # Enable all RP filters
        for interface in ifs:
            if (interface == "all" or interface == "lo" or
                interface in dvb_ifs):
                continue

            # Check current configuration
            current_cfg = _read_filter(interface)

            if (int(current_cfg) > 0):
                if (not non_root):
                    print("RP filter is already enabled on interface %s" %(
                        interface))
            else:
                _add_filter(interface)

        # Disable the overall RP filter
        _rm_filter("all")

        # And disable RP filtering on the DVB interface
        for dvb_if in dvb_ifs:
            _rm_filter(dvb_if)


def set_filters(dvb_ifs, prompt=True):
    """Disable reverse-path (RP) filtering for the DVB interfaces

    Args:
        dvb_ifs : list of DVB network interfaces
        prompt  : Whether to prompt user before applying a configuration

    """
    assert(isinstance(dvb_ifs, list))

    print("\n----------------------------- Reverse Path Filters " +
          "-----------------------------")

    # Check if RP filters are already configured properly
    if (os.geteuid() == 0):
        rp_filters_set = list()
        for dvb_if in dvb_ifs:
            rp_filters_set.append(_check_rp_filters(dvb_if))

        if (all(rp_filters_set)):
            print("Current RP filtering configurations are already OK")
            print("Skipping...")
            return

    util.fill_print("Blocksat traffic is one-way and thus reverse path (RP) \
    filtering must be disabled.")

    # For a non-root user, just run without asking. It is not going to run the
    # command anyway, just print it to stdout.
    non_root = os.geteuid() != 0
    if (non_root):
        util.fill_print("To configure RP filters, run blocksat-cli as root or \
        run the following commands on your own:")
    else:
        util.fill_print(
            "The automatic solution disables RP filtering on the DVB interface \
            and enables RP filtering on all other interfaces.")

    if (non_root or (not prompt) or util._ask_yes_or_no("OK to proceed?")):
        _set_filters(dvb_ifs, non_root)
    else:
        print("RP filtering configuration cancelled")


def subparser(subparsers):
    """Parser for rp command"""
    p = subparsers.add_parser('reverse-path', aliases=['rp'],
                              description="Set reverse path filters",
                              help='Set reverse path filters',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-i', '--interface', required = True,
                   help='Network interface')
    p.set_defaults(func=run)
    return p


def run(args):
    """Call function that sets reverse path filters

    Handles the reverse-path subcommand

    """
    set_filters([args.interface])





"""Manage software dependencies"""
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from . import config, defs, util
import os, subprocess, logging
from shutil import which
from pprint import pformat
logger = logging.getLogger(__name__)


def _install_packages(apt_list, dnf_list):
    """Install binary packages"""
    util._print_sub_header("Binary Dependencies")
    if (which("apt")):
        cmd = ["apt", "install"]
        cmd.extend(apt_list)
    elif (which("dnf")):
        cmd = ["dnf", "install"]
        cmd.extend(dnf_list)
    else:
        raise RuntimeError("Could not find a supported package manager")

    util.run_and_log(util.root_cmd(cmd), logger)


def _install_sdr(srcdir, usrdir, update=False):
    util._print_header("Installing SDR Dependencies")

    bindir = os.path.join(usrdir, "bin")

    if (os.geteuid() != 0):
        util.fill_print("Some commands require root access and will prompt "
                        "for password")

    # Binary packages
    apt_pkg_list = ["git", "make", "g++", "libx11-dev", "gqrx-sdr", "rtl-sdr"]
    dnf_pkg_list = ["git", "make", "g++", "libX11-devel", "gqrx", "rtl-sdr"]
    _install_packages(apt_pkg_list, dnf_pkg_list)

    # leansdr build
    util._print_sub_header("leansdr")
    leansdr_dir     = os.path.join(srcdir, "leansdr")
    leansdr_app_dir = os.path.join(leansdr_dir, "src", "apps")

    is_update = update
    if (not is_update) and os.path.exists(leansdr_dir):
        is_update = True
        print("leansdr build directory already exists. Checking for updates...")
    elif is_update and (not os.path.exists(leansdr_dir)):
        is_update = False
        print("leansdr build directory does no exist. Installing...")

    if (is_update):
        util.run_and_log(["git", "pull", "origin", "master"],
                         logger, cwd=leansdr_dir)
        util.run_and_log(["git", "submodule", "update"],
                         logger, cwd=leansdr_dir)
    else:
        util.run_and_log(["git", "clone", "--recursive",
                          "https://github.com/Blockstream/leansdr.git"],
                         logger, cwd=srcdir)

    nproc = "-j" + subprocess.check_output(["nproc"]).decode().rstrip()

    util.run_and_log(["make", nproc], logger, cwd=leansdr_app_dir)
    util.run_and_log(["install", "leandvb", bindir],
                     logger, cwd=leansdr_app_dir)

    # LDPC
    util._print_sub_header("LDPC Tool")
    ldpc_dir = os.path.join(leansdr_dir, "LDPC")
    util.run_and_log(["make", nproc, "CXX=g++", "ldpc_tool"], logger,
                     cwd=ldpc_dir)
    util.run_and_log(["install", "ldpc_tool", bindir],
                     logger, cwd=ldpc_dir)

    # TSDuck
    util._print_sub_header("TSDuck")

    # Build from source if binary package didn't work
    tsp_dir   = os.path.join(srcdir, "tsduck")
    is_update = update
    if (not is_update) and os.path.exists(tsp_dir):
        is_update = True
        print("tsduck build directory already exists. Checking for updates...")
    elif is_update and (not os.path.exists(tsp_dir)):
        is_update = False
        print("tsduck build directory does no exist. Installing...")

    if (is_update):
        util.run_and_log(["git", "pull", "origin", "master"],
                         logger, cwd=tsp_dir)
    else:
        util.run_and_log(["git", "clone",
                          "https://github.com/tsduck/tsduck.git"],
                         logger, cwd=srcdir)

    util.run_and_log(["build/install-prerequisites.sh"], logger, cwd=tsp_dir)
    util.run_and_log(["make", "NOTELETEXT=1", "NOSRT=1", "NOPCSC=1",
                      "NOCURL=1", "NODTAPI=1"], logger, cwd=tsp_dir)
    util.run_and_log(["make", "install", "SYSPREFIX={}".format(usrdir)], logger,
                     cwd=tsp_dir)


def _print_help(args):
    """Re-create argparse's help menu"""
    parser     = ArgumentParser()
    subparsers = parser.add_subparsers(title='', help='')
    parser     = subparser(subparsers)
    print(parser.format_help())


def subparser(subparsers):
    """Parser for install command"""
    p = subparsers.add_parser('dependencies', aliases=['deps'],
                              description="Manage dependencies",
                              help='Manage dependencies',
                              formatter_class=ArgumentDefaultsHelpFormatter)

    p.set_defaults(func=_print_help)

    subsubp = p.add_subparsers(title='subcommands',
                               help='Target sub-command')

    p1 = subsubp.add_parser('install',
                            description="Install software dependencies",
                            help='Install software dependencies')
    p1.set_defaults(func=run, update=False)

    p1 = subsubp.add_parser('update',
                            description="Update software dependencies",
                            help='Update software dependencies')
    p1.set_defaults(func=run, update=True)

    return p


def run(args):
    """Run installations"""
    info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (info is None):
        return

    srcdir = os.path.join(args.cfg_dir, "src")
    usrdir = os.path.join(args.cfg_dir, "usr")
    bindir = os.path.join(usrdir, "bin")

    if not os.path.exists(srcdir):
        os.makedirs(srcdir)

    if not os.path.exists(usrdir):
        os.makedirs(usrdir)

    if not os.path.exists(bindir):
        os.makedirs(bindir)

    if (info['setup']['type'] == defs.sdr_setup_type):
        _install_sdr(srcdir, usrdir, update=args.update)
    else:
        print("Installation of dependencies no supported yet "
              "for {} demodulator setup".format(info['setup']['type']))





"""Utility functions"""
import os, textwrap, subprocess


def fill_print(text):
    text = " ".join(text.replace("\n", "").split())
    print(textwrap.fill(text))
    print()


def typed_input(msg, hint=None, in_type=int):
    """Ask for user input of a specific type"""
    res = None
    while (res is None):
        try:
            res = in_type(input(msg + ": "))
        except ValueError:
            if (hint is None):
                if (in_type == int):
                    print("Please enter an integer number")
                elif (in_type == float):
                    print("Please enter a number")
                else:
                    type_str = in_type.__name__
                    print("Please enter a {}".format(type_str))
            else:
                print(hint)
    assert(res is not None)
    assert(isinstance(res, in_type))
    return res


def _ask_yes_or_no(msg, default="y", help_msg = None):
    """Yes or no question

    Args:
        msg      : the message or question to ask the user
        default  : default response
        help_msg : Optional help message

    Returns:
        True if answer is yes, False otherwise.

    """
    response = None

    if (default == "y"):
        options = "[Y/n]"
    else:
        options = "[N/y]"

    while response not in {"y", "n"}:
        if (help_msg is None):
            question = msg + " " + options + " "
            raw_resp = input(question) or default
        else:
            print(textwrap.fill(msg))
            print()
            print(textwrap.fill(help_msg))
            print()
            raw_resp = input("Answer " + options + " ") or default

        response = raw_resp.lower()

        if (response not in ["y", "n"]):
            print("Please enter \"y\" or \"n\"")

    return (response == "y")


def _ask_multiple_choice(vec, msg, label, to_str, help_msg = None,
                         none_option = False, none_str = "None of the above"):
    """Multiple choice question

    Args:
        vec         : Vector with elements to choose from
        msg         : Msg to prompt user for choice
        label       : Description/label of what "vec" holdes
        to_str      : Function that prints information about elements
        help_msg    : Optional help message
        none_option : Whether to display a "none of the above" option
        none_str    : What do display as "none of the above" message

    Returns:
        Chosen element

    """
    assert(len(vec) > 1)

    print(msg)

    for i_elem, elem in enumerate(vec):
        elem_str = to_str(elem)
        print("[%2u] %s" %(i_elem, elem_str))

    if (none_option):
        print("[%2u] %s" %(len(vec), none_str))

    if (help_msg is not None):
        print()
        print(help_msg)

    resp = None
    while (not isinstance(resp, int)):
        try:
            resp = int(input("\n%s number: " %(label)))
        except ValueError:
            print("Please choose a number")
            continue

        max_resp = len(vec) + 1 if none_option else len(vec)
        if (resp >= max_resp):
            print("Please choose number from 0 to %u" %(max_resp - 1))
            resp = None
            continue

        if (none_option and resp == len(vec)):
            choice = None
            print(none_str)
        else:
            choice = vec[resp]
            print(to_str(choice))
        print()

        return choice


def _print_header(header, target_len=80):
    """Print section header"""

    prefix      = ""
    suffix      = ""
    header_len  = len(header) + 2
    remaining   = target_len - header_len
    prefix_len  = int(remaining / 2)
    suffix_len  = int(remaining / 2)

    if (remaining % 1 == 1):
        prefix_len += 1

    for i in range(0, prefix_len):
        prefix += "-"

    for i in range(0, suffix_len):
        suffix += "-"

    print("\n" + prefix + " " + header + " " + suffix)


def _print_sub_header(header, target_len=60):
    """Print sub-section header"""
    _print_header(header, target_len=target_len)


def prompt_for_enter():
    try:
        resp = input("\nPress Enter to continue...")
        if (resp == "q"):
            exit()
        os.system('clear')
    except KeyboardInterrupt:
        print("Aborting")
        exit()

def root_cmd(cmd):
    """Add sudo to cmd if non-root

    Args:
        cmd : Command as list

    """
    assert(isinstance(cmd, list))
    if (os.geteuid() != 0 and cmd[0] != "sudo"):
        cmd.insert(0, "sudo")
    return cmd


def run_or_print_root_cmd(cmd, logger=None):
    """Run root command

    Args:
        cmd : Command as list

    """
    assert(isinstance(cmd, list))
    assert(cmd[0] != "sudo")

    if (os.geteuid() == 0):
        if (logger is not None):
            logger.debug("> " + " ".join(cmd))
        return subprocess.check_output(cmd)
    else:
        print("> " + " ".join(cmd) + "\n")


def run_and_log(cmd, logger=None, cwd=None):
    assert(isinstance(cmd, list))
    if (logger is not None):
        logger.debug("> " + " ".join(cmd))
    subprocess.run(cmd, cwd=cwd)





"""SDR Receiver Wrapper"""
from argparse import ArgumentDefaultsHelpFormatter
from . import config, defs, util
import subprocess, logging, textwrap, os
logger = logging.getLogger(__name__)


def _tune_max_pipe_size(pipesize):
    """Tune the maximum size of pipes"""
    try:
        subprocess.check_output(["which", "sysctl"])
    except subprocess.CalledProcessError:
        logging.error("Couldn't tune max-pipe-size. Please check how to tune "
                      "it in your OS.")
        return False

    ret = subprocess.check_output(["sysctl", "fs.pipe-max-size"])
    current_max = int(ret.decode().split()[-1])

    if (current_max < pipesize):
        cmd = util.root_cmd(["sysctl", "-w",
                             "fs.pipe-max-size=" + str(pipesize)])

        print(textwrap.fill("The maximum pipe size that is currently "
                            "configured in your OS is of {} bytes, which is "
                            "not sufficient for the demodulator application. "
                            "It will be necessary to run the following command "
                            "as root:".format(current_max), width=80))
        print("\n" + " ".join(cmd) + "\n")

        if (not util._ask_yes_or_no("Is that OK?", default="y")):
            print("Abort")
            return False

        return subprocess.check_output(cmd)
    else:
        return True


def _check_apps(tsp_disabled, bindir):
    """Check if required apps are installed"""
    try:
        subprocess.check_output(["which", "rtl_sdr"])
    except subprocess.CalledProcessError:
        logging.error("Couldn't find rtl_sdr. Is it installed?")
        return False

    if (not os.path.isfile(os.path.join(bindir, "leandvb"))):
        logging.error("Couldn't find leandvb. Is it installed?")
        return False

    if (not os.path.isfile(os.path.join(bindir, "ldpc_tool"))):
        logging.error("Couldn't find ldpc_tool. Is it installed?")
        return False

    if (tsp_disabled):
        return True

    if (not os.path.isfile(os.path.join(bindir, "tsp"))):
        logging.error("Couldn't find tsp. Is it installed?")
        return False

    return True


def subparser(subparsers):
    """Parser for sdr command"""
    p = subparsers.add_parser('sdr',
                              description="Launch SDR receiver",
                              help='Launch SDR receiver',
                              formatter_class=ArgumentDefaultsHelpFormatter)

    rtl_p = p.add_argument_group('rtl_sdr options')
    rtl_p.add_argument('--sps', default=2.0, type=float,
                       help='Samples per symbol, or, equivalently, the '
                           'target oversampling ratio')
    rtl_group = rtl_p.add_mutually_exclusive_group()
    rtl_group.add_argument('-g', '--gain', default=40, type=float,
                           help='RTL-SDR Rx gain')
    rtl_group.add_argument('-f', '--iq-file', default=None,
                           help='File to read IQ samples from instead of reading '
                           'from the RTL-SDR in real-time')

    ldvb_p = p.add_argument_group('leandvb options')
    ldvb_p.add_argument('-n', '--n-helpers', default=6, type=int,
                        help='Number of instances of the external LDPC decoder \
                        to spawn as child processes')
    ldvb_p.add_argument('-d', '--debug-ts', action='count', default=0,
                        help="Activate debugging on leandvb. Use it multiple "
                        "times to increase the debugging level")
    ldvb_p.add_argument('-v', '--verbose', default=False, action='store_true',
                        help='leandvb in verbose mode')
    ldvb_p.add_argument('--gui', default=False, action='store_true',
                        help='GUI mode')
    ldvb_p.add_argument('--derotate', default=0, type=float,
                        help='Frequency offset correction to apply in kHz')
    ldvb_p.add_argument('--fastlock', default=False, action='store_true',
                        help='leandvb fast lock mode')
    ldvb_p.add_argument('--rrc-rej', default=30, type=int,
                        help='leandvb RRC rej parameter')
    ldvb_p.add_argument('-m', '--modcod', default="low",
                        choices=["low", "high"],
                        help="Choose low-throughput vs high-throughput MODCOD")
    ldvb_p.add_argument('--ldpc-dec', default="ext", choices=["int", "ext"],
                        help="LDPC decoder to use (internal or external)")
    ldvb_p.add_argument('--ldpc-bf', default=100,
                        help='Max number of iterations used by the internal \
                        LDPC decoder when not using an external LDPC tool')
    ldvb_p.add_argument('--ldpc-iterations', default=25,
                        help='Max number of iterations used by the external \
                        LDPC decoder when using an external LDPC tool')
    ldvb_p.add_argument('--framesizes', type=int, default=1, choices=[0,1,2,3],
                        help="Bitmask of desired frame sizes (1=normal, \
                        2=short)")
    ldvb_p.add_argument('--no-tsp', default=False, action='store_true',
                        help='Feed leandvb output to stdout instead of tsp')
    ldvb_p.add_argument('--pipe-size', default=32, type=int,
                        help='Size in Mbytes of the input pipe file read by \
                        leandvb')

    tsp_p = p.add_argument_group('tsduck options')
    tsp_p.add_argument('--buffer-size-mb', default=1.0, type=float,
                       help='Input buffer size in MB')
    tsp_p.add_argument('--max-flushed-packets', default=10, type=int,
                       help='Maximum number of packets processed by a tsp '
                       'processor')
    tsp_p.add_argument('--max-input-packets', default=10, type=int,
                       help='Maximum number of packets received at a time from '
                       'the tsp input plugin ')
    tsp_p.add_argument('-p', '--bitrate-period', default=1, type=int,
                       help='Period of bitrate reports in seconds')
    tsp_p.add_argument('-l', '--local-address', default="127.0.0.1",
                       help='IP address of the local interface on which to '
                       'listen for UDP datagrams')
    tsp_p.add_argument('-a', '--analyze', default=False, action='store_true',
                       help='Analyze transport stream and save report on '
                       'program termination')
    tsp_p.add_argument('--analyze-file', default="ts-analysis.txt",
                       action='store_true',
                       help='File on which to save the MPEG-TS analysis.')
    tsp_p.add_argument('--no-bitrate-monitoring', default=False,
                       action='store_true',
                       help='Disable bitrate monitoring')
    tsp_p.add_argument('--monitor-ts', default=False,
                       action='store_true',
                       help='Mnitor MPEG TS sequence discontinuities')
    p.set_defaults(func=run,
                   record=False)

    subsubparsers = p.add_subparsers(title='subcommands',
                                     help='Target SDR sub-command')
    # IQ recording
    p2 = subsubparsers.add_parser('rec',
                                  description="Record IQ samples instead of "
                                  "feeding them into leandvb",
                                  help='Record IQ samples',
                                  formatter_class=ArgumentDefaultsHelpFormatter)
    p2.add_argument('-f', '--iq-file', default="blocksat.iq",
                    help='File on which to save IQ samples received with '
                    'the RTL-SDR.')
    p2.set_defaults(record=True)

    return p


def run(args):
    info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (info is None):
        return

    bindir = os.path.join(args.cfg_dir, "usr", "bin")

    pipe_size_bytes = int(args.pipe_size * (2**20))
    if (not _tune_max_pipe_size(pipe_size_bytes)):
        return

    if (not _check_apps((args.no_tsp or args.record), bindir)):
        print("\nTo install software dependencies, run:")
        print("""
        blocksat-cli deps install
        """)
        return

    # Demodulator configs
    l_band_freq = info['freqs']['l_band']*1e6
    modcod      = defs.low_rate_modcod if args.modcod == "low" else \
                  defs.high_rate_modcod
    sym_rate    = defs.sym_rate[info['sat']['alias']]
    samp_rate   = args.sps*sym_rate

    assert(samp_rate < 2.4e6), \
        "Sample rate of {} exceeds the RTL-SDR limit".format(samp_rate)

    # Derotate up to the CFO recovery range on leandvb. If the desired
    # derotation exceeds the CFO correction range, change the L-band frequency
    # directly on the RTL instead.
    cfo_corr_range = (sym_rate / 4)
    derotate       = args.derotate*1e3
    while (derotate > cfo_corr_range):
        derotate    -= cfo_corr_range
        l_band_freq += cfo_corr_range
    while (derotate < -cfo_corr_range):
        derotate    += cfo_corr_range
        l_band_freq -= cfo_corr_range

    if (args.iq_file is None or args.record):
        rtl_cmd = ["rtl_sdr", "-g", str(args.gain), "-f",
                   str(l_band_freq), "-s", str(samp_rate)]
        if (args.record):
            bytes_per_sec = (2*samp_rate)/(2**20)
            print("IQ recording will be saved on file {}".format(
                args.iq_file))
            print(textwrap.fill("NOTE: the file will grow by approximately "
                                "{} MB per second.".format(bytes_per_sec)))
            if (not util._ask_yes_or_no("Proceed?", default="y")):
                return
            rtl_cmd.append(args.iq_file)
        else:
            rtl_cmd.append("-")

    ldvb_path = os.path.join(bindir, "leandvb")
    ldvb_cmd  = [ldvb_path, "--nhelpers", str(args.n_helpers), "-f",
                str(samp_rate), "--sr", str(sym_rate), "--roll-off",
                str(defs.rolloff), "--standard", "DVB-S2", "--sampler", "rrc",
                "--rrc-rej", str(args.rrc_rej), "--modcods", modcod,
                "--framesizes", str(args.framesizes), "--inpipe",
                str(pipe_size_bytes)]
    if (args.debug_ts == 1):
        ldvb_cmd.append("-d")
    elif (args.debug_ts > 1):
        ldvb_cmd.extend(["-d", "-d"])
    if (args.ldpc_dec == "ext"):
        ldpc_tool_path = os.path.join(bindir, "ldpc_tool")
        ldvb_cmd.extend(["--ldpc-helper", ldpc_tool_path,
                         "--ldpc-iterations", str(args.ldpc_iterations)])
    else:
        ldvb_cmd.extend(["--ldpc-bf", str(args.ldpc_bf)])
    if (args.gui):
        ldvb_cmd.append("--gui")
    if (args.verbose):
        ldvb_cmd.append("-v")
    if (args.fastlock):
        ldvb_cmd.append("--fastlock")
    if (derotate != 0.0):
        ldvb_cmd.extend(["--derotate", str(int(derotate))])

    ldvb_stderr = None if (args.debug_ts > 0) else subprocess.DEVNULL

    # Input
    tsp_path = os.path.join(bindir, "tsp")
    tsp_cmd  = [tsp_path, "--realtime", "--buffer-size-mb",
                str(args.buffer_size_mb), "--max-flushed-packets",
                str(args.max_flushed_packets), "--max-input-packets",
                str(args.max_input_packets)]
    # MPEG-TS Analyzer
    if (args.analyze):
        print("MPEG-TS analysis will be saved on file {}".format(
            args.analyze_file))
        if (not util._ask_yes_or_no("Proceed?", default="y")):
            return
        tsp_cmd.extend(["-P", "analyze", "-o", args.analyze_file])
    if (not args.no_bitrate_monitoring):
        # Monitor Bitrate
        tsp_cmd.extend(["-P", "bitrate_monitor", "-p",
                        str(args.bitrate_period), "--min", "0"])
    if (args.monitor_ts):
        # Monitor MPEG TS discontinuities
        tsp_cmd.extend(["-P", "continuity"])

    # MPE plugin
    tsp_cmd.extend(["-P", "mpe", "--pid",
                    "-".join([str(pid) for pid in defs.pids]), "--udp-forward",
                    "--local-address", args.local_address])
    # Output
    tsp_cmd.extend(["-O", "drop"])

    logger.debug("Run:")

    if (args.record):
        p1 = subprocess.Popen(rtl_cmd)
        try:
            p1.communicate()
        except KeyboardInterrupt:
            p1.kill()
        return
    elif (args.iq_file is None):
        full_cmd  = "> " + " ".join(rtl_cmd) + " | \\\n" + \
                    " ".join(ldvb_cmd)
        p1 = subprocess.Popen(rtl_cmd, stdout=subprocess.PIPE)
        p2 = subprocess.Popen(ldvb_cmd, stdin=p1.stdout, stdout=subprocess.PIPE,
                              stderr=ldvb_stderr)
    else:
        full_cmd   = "> " + " ".join(ldvb_cmd) + " < " + args.iq_file
        fd_iq_file = open(args.iq_file)
        p2 = subprocess.Popen(ldvb_cmd, stdin=fd_iq_file,
                              stdout=subprocess.PIPE,
                              stderr=ldvb_stderr)
    if (not args.no_tsp):
        full_cmd += " | \\\n" + " ".join(tsp_cmd)
        logger.debug(full_cmd)
        p3 = subprocess.Popen(tsp_cmd, stdin=p2.stdout)
        try:
            p3.communicate()
        except KeyboardInterrupt:
            p3.kill()
            p2.kill()
            p1.kill()
    else:
        logger.debug(full_cmd)
        try:
            p2.communicate()
        except KeyboardInterrupt:
            p2.kill()
            p1.kill()





"""Linux USB demodulator"""
from pprint import pformat
from ipaddress import IPv4Interface
import os, sys, signal, argparse, subprocess, time, logging, threading, json
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from . import config, util, defs, rp, firewall, ip
import textwrap
from shutil import which
logger = logging.getLogger(__name__)


def _setup_logfile(cfg_dir):
    """Setup directory and file for dvbv5-zap logs"""
    log_dir = os.path.join(cfg_dir, "logs")
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    name    = "usb-" + time.strftime("%Y%m%d-%H%M%S") + ".log"
    logfile = os.path.join(log_dir, name)
    return logfile


def _find_v4l_lnb(info):
    """Find suitable LNB within v4l-utils preset LNBs

    The LNBs on list v4l_lnbs are defined at lib/libdvbv5/dvb-sat.c

    """


    target_lo_freq_raw  = info['lnb']['lo_freq']
    target_is_universal = info['lnb']['universal']

    if (target_is_universal):
        assert(isinstance(target_lo_freq_raw, list))
        assert(len(target_lo_freq_raw) == 2)
        target_lo_freq = [int(x) for x in target_lo_freq_raw]
    else:
        target_lo_freq = [int(target_lo_freq_raw)]

    # Find options that match the LO freq
    options = list()
    for lnb in defs.v4l_lnbs:
        is_universal_option = 'highfreq' in lnb

        if (target_is_universal and (not is_universal_option)):
            continue

        if (info['lnb']['pol'].lower() == "dual" and # user has dual-pol LNB
            (info['sat']['pol'] == "H" or # and satellite requires H pol
             (info['lnb']['v1_pointed'] and # or LNB already operates with H pol
              info['lnb']['v1_psu_voltage'] >= 16)
            ) and "rangeswitch" not in lnb): # but LNB candidate is single-pol
            continue # not a validate candidate, skip

        if (target_is_universal):
            if (is_universal_option and
                lnb['lowfreq'] == target_lo_freq[0] and
                lnb['highfreq'] == target_lo_freq[1]):
                options.append(lnb)
        else:
            if (lnb['lowfreq'] == target_lo_freq[0]):
                options.append(lnb)

    assert(len(options) > 0), "LNB doesn't match a valid option"
    logger.debug("Matching LNB options: {}".format(pformat(options)))

    return options[0]


def _find_adapter(list_only=False):
    """Find the DVB adapter

    Returns:
        Tuple with (adapter index, frontend index)

    """
    util._print_header("Find DVB Adapter")
    ps     = subprocess.Popen("dmesg", stdout=subprocess.PIPE)
    try:
        output = subprocess.check_output(["grep", "frontend"], stdin=ps.stdout,
                                         stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as grepexc:
        if (grepexc.returncode == 1):
            output = ""
            pass
    ps.wait()

    # Search a range of adapters. There is no command to list all adapters, so
    # we try to list each one individually using `dvbnet -a adapter_no -l`.
    adapters = list()
    for a in range(0,10):
        cmd     = ["dvbnet", "-a", str(a), "-l"]
        logger.debug("> " + " ".join(cmd))

        with open(os.devnull, 'w') as devnull:
            res = subprocess.call(cmd, stdout=devnull, stderr=devnull)
            if (res == 0):
                # Try a few frontends too
                for f in range(0,2):
                    try:
                        output = subprocess.check_output(["dvb-fe-tool", "-a",
                                                          str(a), "-f", str(f)])
                        line   = output.splitlines()[0].decode().split()
                        adapter = {
                            "adapter"  : str(a),
                            "frontend" : line[5].replace(")","").split("frontend")[-1],
                            "vendor"   : line[1],
                            "model"    : " ".join(line[2:4]),
                            "support"  : line[4]
                        }
                        adapters.append(adapter)
                        logger.debug(pformat(adapter))
                    except subprocess.CalledProcessError as e:
                        pass

    # If nothing was obtained using dvbnet, try to inspect dmesg logs
    if (len(adapters) == 0):
        lines       = output.splitlines()
        adapter_set = set() # use set to filter unique values
        adapters    = list()
        for line in lines:
            linesplit = line.decode().split()
            if ("adapter" not in linesplit):
                continue
            i_adapter  = linesplit.index('adapter')
            i_frontend = linesplit.index('frontend')
            device     = linesplit[i_frontend + 2:]
            adapter = {
                "adapter"  : linesplit[i_adapter + 1],
                "frontend" : linesplit[i_frontend + 1],
                "vendor"   : device[0][1:],
                "model"    : " ".join(device[1:-1]),
                "support"  : device[-1][:-4].replace('(', '').replace(')', '')
            }

            # Maybe the device on dmesg does not exist anymore. Check!
            try:
                cmd = ["dvbnet", "-a", adapter['adapter'], "-l"]
                logger.debug("> " + " ".join(cmd))
                res = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
            except subprocess.CalledProcessError as e:
                continue

            # If it exists, add to set of candidate adapters
            adapter_set.add(json.dumps(adapter))

        # Process unique adapter logs
        for adapter in adapter_set:
            adapters.append(json.loads(adapter))
            logger.debug(pformat(json.loads(adapter)))

    dvb_s2_adapters = [a for a in adapters if ("DVB-S/S2" in a["support"])]
    logger.debug(dvb_s2_adapters)

    assert(len(dvb_s2_adapters) > 0), "No DVB-S2 adapters found"

    chosen_adapter = None
    for adapter in dvb_s2_adapters:
        print("Found DVB-S2 adapter: %s %s" %(adapter["vendor"],
                                              adapter["model"]))

        if (not list_only):
            if (len(dvb_s2_adapters) == 1 or
                util._ask_yes_or_no("Choose adapter?")):
                chosen_adapter = adapter
                logger.debug("Chosen adapter:")
                logger.debug(pformat(adapter))
                break

    if (list_only):
        return

    if (chosen_adapter is None):
        raise ValueError("Please choose DVB-S2 adapter")

    return chosen_adapter["adapter"], chosen_adapter["frontend"]


def _dvbnet_single(adapter, ifname, pid, ule, existing_dvbnet_interfaces):
    """Start DVB network interface

    Args:
        adapter                    : DVB adapter index
        ifname                     : DVB network interface name
        pid                        : PID to listen to
        ule                        : Whether to use ULE framing
        existing_dvbnet_interfaces : List of dvbnet interfaces already
                                     configured for the adapter

    """

    assert(pid >= 32 and pid <= 8190), "PID not insider range 32 to 8190"

    if (ule):
        encapsulation = 'ULE'
    else:
        encapsulation = 'MPE'

    # Check if interface already exists
    res = subprocess.call(["ip", "addr", "show", "dev", ifname],
                          stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    os_interface_exists = (res == 0)
    matching_dvbnet_if  = None

    # When the network interface exists in the OS, we also need to check if the
    # matching dvbnet device is configured according to what we want now
    if (os_interface_exists):
        print("Network interface %s already exists" %(ifname))

        for interface in existing_dvbnet_interfaces:
            if (interface['name'] == ifname):
                matching_dvbnet_if = interface
                break

    # Our indication that interface exists comes from "ip addr show
    # dev". However, it is possible that dvbnet does not have any interface
    # associated to an adapter, so check if we found anything:
    cfg_interface = False
    if (len(existing_dvbnet_interfaces) > 0 and matching_dvbnet_if is not None):
        # Compare to desired configurations
        if (matching_dvbnet_if['pid'] != pid or
            matching_dvbnet_if['encapsulation'] != encapsulation):
            cfg_interface = True

        if (matching_dvbnet_if['pid'] != pid):
            print("Current PID is %d. Set it to %d" %(
                matching_dvbnet_if['pid'], pid))

        if (matching_dvbnet_if['encapsulation'] != encapsulation):
            print("Current encapsulation is %s. Set it to %s" %(
                matching_dvbnet_if['encapsulation'], encapsulation
            ))
    else:
        cfg_interface = True

    # Create interface in case it doesn't exist or needs to be re-created
    if (cfg_interface):
        # If interface exists, but must be re-created, remove the existing one
        # first
        if (os_interface_exists):
            _rm_dvbnet_interface(adapter, ifname, verbose=False)

        adapter_dir = '/dev/dvb/adapter' + adapter

        ule_arg = "-U" if ule else ""

        if (os.geteuid() == 0):
            print("Launch {} using {} encapsulation:".format(
                ifname, "ULE" if ule else "MPE"))

        res = util.run_or_print_root_cmd(["dvbnet", "-a", adapter, "-p",
                                          str(pid), ule_arg], logger)
        if (res is not None):
            print(res.decode())
    else:
        print("Network interface %s already configured correctly" %(ifname))


def _dvbnet(adapter, ifnames, pids, ule=False):
    """Start DVB network interfaces of a DVB adapter

    An adapter can have multiple dvbnet interfaces, one for each PID.

    Args:
        adapter  : DVB adapter index
        ifnames  : list of DVB network interface names
        pids     : List of PIDs to listen to on each interface
        ule      : Whether to use ULE framing

    """
    assert(isinstance(ifnames, list))
    assert(isinstance(pids, list))
    assert(len(ifnames) == len(pids)), \
        "Interface names and PID number must be vectors of the same length"

    # Find the dvbnet interfaces that already exist for the chosen adapter
    existing_dvbnet_iif = _find_dvbnet_interfaces(adapter)

    util._print_header("Network Interface")

    if (os.geteuid() != 0):
        util.fill_print("Launch blocksat-cli as root or run the following \
        commands on your own:")

    for ifname, pid in zip(ifnames, pids):
        _dvbnet_single(adapter, ifname, pid, ule, existing_dvbnet_iif)


def _find_dvbnet_interfaces(adapter):
    """Find dvbnet interface(s) of a DVB adapter

    An adapter can have multiple dvbnet interfaces, one for each PID.

    Args:
        adapter: Corresponding DVB adapter

    Returns:
        interfaces : List of dvbnet interfaces

    """

    util._print_header("Find dvbnet interface(s)")
    cmd     = ["dvbnet", "-a", adapter, "-l"]
    logger.debug("> " + " ".join(cmd))
    res     = subprocess.check_output(cmd)

    interfaces = list()
    for line in res.splitlines():
        if ("Found device" in line.decode()):
            line_split    = line.decode().split()
            interface = {
                'dev'           : line_split[2][:-1],
                'name'          : line_split[4][:-1],
                'pid'           : int(line_split[8][:-1]),
                'encapsulation' : line_split[10]
            }
            logger.debug(pformat(interface))
            interfaces.append(interface)

    if (len(interfaces) == 0):
        print("Could not find any dvbnet interface")

    return interfaces


def _rm_dvbnet_interface(adapter, ifname, verbose=True):
    """Remove DVB net interface

    Args:
        adapter   : Corresponding DVB adapter
        interface : dvbnet interface number
        verbose   : Controls verbosity
    """

    if (verbose):
        util._print_header("Remove dvbnet interface")
    res       = util.run_or_print_root_cmd(["ip", "link", "set",
                                            ifname, "down"], logger)
    if_number = ifname.split("_")[-1]
    res       = util.run_or_print_root_cmd(["dvbnet", "-a", adapter, "-d",
                                            if_number], logger)
    print(res.decode())

    # Remove also the static IP address configuration
    ip.rm_ip(ifname)


def zap(adapter, frontend, ch_conf_file, user_info, lnb="UNIVERSAL",
        output=None, timeout=None, monitor=False, scrolling=False):
    """Run zapper

    Args:
        adapter      : DVB adapter index
        frontend     : frontend
        ch_conf_file : Path to channel configurations file
        user_info    : Dictionary with user configurations
        lnb          : LNB type
        output       : Output filename (when recording)
        timeout      : Run the zap for this specified duration
        monitor      : Monitor mode. Monitors DVB traffic stats (throughput and
                       packets per second), but does not deliver data upstream.
        scrolling    : Whether to print zap logs by scrolling rather than
                       printing always on the same line.

    Returns:
        Subprocess object

    """

    util._print_header("Tuning DVB Receiver")
    print("Running dvbv5-zap")

    # LNB name to use when calling dvbv5-zap
    if (lnb is None):
        # Find suitable LNB within v4l-utils preset LNBs
        lnb = _find_v4l_lnb(user_info)['alias']

    cmd = ["dvbv5-zap", "-c", ch_conf_file, "-a", str(adapter), "-f",
           str(frontend), "-l", lnb, "-v"]

    if (output is not None):
        cmd = cmd + ["-o", output]

        if (os.path.exists(output)):
            print("File %s already exists" %(output))

            # The recording is such that MPEG TS packets are overwritten one by
            # one. For instance, if previous ts file had 1000 MPEG TS packets,
            # when overwriting, the tool would overwrite each packet
            # individually. So if it was stopped for instance after the first 10
            # MPEG TS packets, only the first 10 would be overwritten, the
            # remaining MPEG TS packets would remain in the ts file.
            #
            # The other option is to remove the ts file completely and start a
            # new one. This way, all previous ts packets are guaranteed to be
            # erased.
            raw_resp = input("Remove and start new (R) or Overwrite (O)? [R/O] ")
            response = raw_resp.lower()

            if (response == "r"):
                os.remove(output)
            elif (response != "o"):
                raise ValueError("Unknown response")

    if (timeout is not None):
        cmd = cmd + ["-t", timeout]

    if (monitor):
        assert(not scrolling), \
            "Monitor mode does not work with scrolling (line-by-line) logs"
        assert(output is None), \
            "Monitor mode does not work if recording (i.e. w/ -r flag)"
        cmd.append("-m")

    cmd.append("blocksat-ch")

    logger.debug("> " + " ".join(cmd))

    if (scrolling):
        ps = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE, universal_newlines=True)
    else:
        ps = subprocess.Popen(cmd)

    return ps


def subparser(subparsers):
    """Subparser for usb command"""
    p = subparsers.add_parser('usb',
                              description="USB DVB-S2 receiver manager",
                              help='Manage Linux USB DVB-S2 receiver',
                              formatter_class=ArgumentDefaultsHelpFormatter)

    # Common parameters
    p.add_argument('-a', '--adapter',
                   default=None,
                   help='DVB-S2 adapter number')

    p.add_argument('-f', '--frontend',
                   default=None,
                   help='DVB-S2 adapter\'s frontend number')

    p.set_defaults(func=print_help)

    subsubparsers = p.add_subparsers(title='subcommands',
                                         help='Target USB sub-command')

    # Launch
    p1 = subsubparsers.add_parser('launch',
                                  description="Launch a USB DVB-S2 receiver",
                                  help='Launch a Linux USB DVB-S2 receiver')

    p1.add_argument('-c', '--chan-conf',
                    default="channels.conf",
                    help='Channel configurations file within the configuration \
                    directory')

    p1.add_argument('-l', '--lnb',
                    choices=defs.lnb_options,
                    default=None,
                    help="LNB from v4l-utils to be used. "
                    "If None, i.e. not specified, it will be set "
                    "automatically")

    p1.add_argument('-r', '--record-file', default=None,
                    help='Record MPEG-TS traffic into target file')

    p1.add_argument('-t', '--timeout', default=None,
                    help='Stop zapping after timeout - useful to \
                    control recording time')

    p1.add_argument('-m', '--monitor', default=False,
                    action='store_true',
                    help='Launch dvbv5-zap in monitor mode - useful \
                    to debug packet and bit rates')

    p1.add_argument('-s', '--scrolling', default=False,
                    action='store_true',
                    help='Print dvbv5-zap logs line-by-line, i.e. \
                    scrolling, rather than always on the same line')

    p1.add_argument('--logfile', default=False,
                    action='store_true',
                    help='Save dvbv5-zap logs on a file')

    p1.set_defaults(func=launch)

    # Initial configurations
    p2 = subsubparsers.add_parser('config', aliases=['cfg'],
                                  description='Initial configurations',
                                  help='Configure DVB-S2 interface(s) and\
                                  the host')
    p2.add_argument('-U', '--ule', default=False,
                    action='store_true',
                    help='Use ULE encapsulation instead of MPE')

    p2.add_argument('--skip-rp', default=False, action='store_true',
                    help='Skip settting of reverse path filters')

    p2.add_argument('--skip-firewall', default=False,
                    action='store_true',
                    help='Skip configuration of firewall rules')

    p2.add_argument('--pid', default=defs.pids,
                    type=int,
                    nargs='+',
                    help='List of PIDs to be listened to by dvbnet')

    p2.add_argument('-i', '--ip',
                    default=None,
                    nargs='+',
                    help='IP address set for each DVB-S2 net \
                    interface with subnet mask in CIDR notation')

    p2.add_argument('-y', '--yes', default=False, action='store_true',
                    help="Default to answering Yes to configuration prompts")

    p2.set_defaults(func=usb_config)

    # Find adapter sub-command
    p3 = subsubparsers.add_parser('list', aliases=['ls'],
                                  description="List DVB-S2 adapters",
                                  help='List DVB-S2 adapters',
                                  formatter_class=ArgumentDefaultsHelpFormatter)
    p3.set_defaults(func=list_subcommand)

    # Remove adapter sub-command
    p4 = subsubparsers.add_parser('remove', aliases=['rm'],
                                  description="Remove DVB-S2 adapter",
                                  help='Remove DVB-S2 adapter',
                                  formatter_class=ArgumentDefaultsHelpFormatter)
    p4.add_argument('-a', '--adapter',
                           default=None,
                           help='DVB-S2 adapter number')
    p4.set_defaults(func=rm_subcommand)

    return p


def _common(args):
    # User info
    user_info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (user_info is None):
        return

    # Find adapter
    if (args.adapter is None):
        adapter, frontend = _find_adapter()
    else:
        adapter  = args.adapter
        frontend = args.frontend

    return user_info, adapter, frontend


def usb_config(args):
    """Config the DVB interface(s) and the host"""
    if (os.geteuid() != 0):
        print("WARNING:\n")
        util.fill_print("\"blocksat-cli usb config\" requires root access. \
        Please run as root or follow the instructions below.")

    common_params = _common(args)
    if (common_params is None):
        return
    user_info, adapter, frontend = common_params

    if args.ip is None:
        ips = ip.compute_rx_ips(user_info['sat']['ip'], len(args.pid))
    else:
        ips = args.ip

    assert(all(["/" in x for x in ips])), "Please provide IPs in CIDR notation"

    assert(len(args.pid) == len(ips)), \
        "Number of PIDs (%u) defined by argument --pid " %(len(args.pid)) + \
        "does not match the number of IPs (%u) defined by " %(len(ips)) + \
        "argument --ip. Please define one IP address for each PID."

    # dvbnet interfaces of interest
    #
    # NOTE: there is one dvbnet interface per PID. Each interface will have a
    # different IP address.
    net_ifs  = list()
    for i_device in range(0, len(args.pid)):
        # Define interface name that is going to be generated by dvbnet
        net_if = "dvb" + adapter + "_" + str(i_device)
        net_ifs.append(net_if)

    # Create the dvbnet interface(s)
    _dvbnet(adapter, net_ifs, args.pid, ule=args.ule)

    # Set RP filters
    if (not args.skip_rp):
        rp.set_filters(net_ifs, prompt=(not args.yes))

    # Set firewall rules
    if (not args.skip_firewall):
        firewall.configure(net_ifs, defs.src_ports, prompt=(not args.yes))

    # Set IP
    ip.set_ips(net_ifs, ips)

    util._print_header("Next Step")
    print("Run:\n\nblocksat-cli usb launch\n")

    return


def launch(args):
    """Launch the DVB interface from scratch

    Handles the launch subcommand

    """

    common_params = _common(args)
    if (common_params is None):
        return
    user_info, adapter, frontend = common_params

    # Channel configuration file
    chan_conf = os.path.join(args.cfg_dir, os.path.basename(args.chan_conf))

    # Zap
    zap_ps = zap(adapter, frontend, chan_conf, user_info, lnb=args.lnb,
                 output=args.record_file, timeout=args.timeout,
                 monitor=args.monitor, scrolling=(args.scrolling
                                                  or args.logfile))

    # Prepare logs
    logfile = None if (not args.logfile) else _setup_logfile(args.cfg_dir)

    # Handler for SIGINT
    def signal_handler(sig, frame):
        print('Stopping...')
        zap_ps.terminate()
        sys.exit(zap_ps.poll())

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    # Listen to dvbv5-zap indefinitely
    if (args.scrolling or args.logfile):
        # Loop indefinitely over zap
        prev_line = None
        while (zap_ps.poll() is None):
            line = zap_ps.stderr.readline()
            if (line and line != "\n"):
                pretty_line = '\r{}: '.format(
                    time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())) + line
                if (("Signal" in line) and ("Layer" not in line)):
                    prev_line = pretty_line.replace("\n", " ")
                else:
                    concat_line = pretty_line if prev_line is None else \
                                  (prev_line + line)
                    final_line  = " ".join(concat_line.split()) + "\n"
                    prev_line   = None
                    print(final_line, end='')
                    if (logfile is not None):
                        with open(logfile, 'a') as fd:
                            fd.write(final_line)
            else:
                time.sleep(1)
    else:
        zap_ps.wait()
    sys.exit(zap_ps.poll())


def list_subcommand(args):
    """Call function that finds the DVB adapter

    Handles the find-adapter subcommand

    """
    _find_adapter(list_only=True)


def rm_subcommand(args):
    """Remove DVB interface

    """

    if (os.geteuid() != 0):
        logger.error("Root access is required to remove dvbnet interfaces.")
        print("Please, run as root.")
        return

    # Find adapter
    if (args.adapter is None):
        adapter, frontend = _find_adapter()
    else:
        adapter = args.adapter

    interfaces     = _find_dvbnet_interfaces(adapter)
    chosen_devices = list()

    if (len(interfaces) > 1):
        print("Choose net device to remove:")
        for i_dev, interface in enumerate(interfaces):
            print("[%2u] %s" %(i_dev, interface['name']))
        print("[ *] all")

        try:
            choice = input("Choose number: ")
            if (choice == "*"):
                i_chosen_devices = range(0, len(interfaces))
            else:
                i_chosen_devices = [int(choice)]
        except ValueError:
            raise ValueError("Please choose a number or \"*\" for all devices")

        for i_chosen_dev in i_chosen_devices:
            if (i_chosen_dev > len(interfaces)):
                raise ValueError("Invalid number")

            chosen_devices.append(interfaces[i_chosen_dev]['name'])

    elif (len(interfaces) == 0):
        print("No DVB network interfaces to remove")
        return
    else:
        # There is a single interface
        chosen_devices.append(interfaces[0]['name'])

    for chosen_dev in chosen_devices:
        if (not util._ask_yes_or_no("Remove interface %s?" %(chosen_dev))):
            print("Aborting...")
            return

        _rm_dvbnet_interface(adapter, chosen_dev)


def print_help(args):
    """Re-create argparse's help menu for the usb command"""
    parser     = ArgumentParser()
    subparsers = parser.add_subparsers(title='', help='')
    parser     = subparser(subparsers)
    print(parser.format_help())





"""Instructions for the user"""
from argparse import ArgumentDefaultsHelpFormatter
from . import util, defs, config
import textwrap, logging


def _item(text):
    print(textwrap.fill(text, initial_indent="- ", subsequent_indent="  "))


def _print(text):
    text = " ".join(text.replace("\n", "").split())
    print(textwrap.fill(text))
    print()


def _print_s400_instructions(info):
    """Print instructions for configuration of the Novra S400
    """
    util._print_header("Novra S400")

    _print("""
    The Novra S400 is a standalone demodulator, which will receive data from
    satellite and output IP packets to the host over the network. Hence, you will
    need to configure both the S400 and the host.
    """)

    util._print_sub_header("Connections")

    _print("The Novra S400 can be connected as follows:")


    print(("LNB ----> S400 (RF1 Interface) -- "
           "S400 (LAN 1 Interface) ----> Host / Network\n"))


    _item("Connect the LNB directly to interface RF1 of the S400 using a "
          "coaxial cable (an RG6 cable is recommended).")
    _item("Connect the S400's LAN1 interface to your computer or network.")

    util.prompt_for_enter()

    util._print_sub_header("S400's web user interface (UI)")
    print("Next, you need to access the web UI of the S400:\n")
    _item(
        "Configure you host's network interface to the same subnet as the S400"
        ". By default, the S400 is configured with IP address 192.168.1.2 on "
        "LAN1 and 192.168.2.2 on LAN2. So, if you are connecting to LAN1, "
        "make sure your host's network interface has IP address 192.168.1.x, "
        "where \"x\" could be any number higher than 2. For example, you could "
        "configure your host's network interface with IP address 192.168.1.3.")
    _item("From your browser, access 192.168.1.2 (or 192.168.2.2 if "
          "connected to LAN 2).")
    _item("The web management console should open.")
    print()

    util.prompt_for_enter()

    util._print_sub_header("S400 FW Version")
    print("In the web UI, go to System > About:")
    print("Confirm that the version of the Configuration Agent is 1.6.1 or higher.")
    print()

    util.prompt_for_enter()

    util._print_sub_header("S400 Configurations")
    print("1. First you need to log in as admin, on the top right of the page.")
    _item("Password: \"password\"")
    print()

    print("2. Go to Interfaces > RF1 and configure as follows:\n")
    _item("DVB Mode: \"DVB-S2\"")
    _item("LBand: {:.1f} MHz".format(info['freqs']['l_band']))
    _item("Symbol Rate: {} MBaud".format(
        defs.sym_rate[info['sat']['alias']]/1e6))
    _item("MODCOD: VCM")
    _item("Gold Code: 0")
    _item("Input Stream ID: 0")
    _item("LNB Power On: Enable")
    _item("L.O. Frequencies: {:.1f} MHz".format(info['freqs']['lo']))

    if (info['lnb']['pol'].lower() == "dual" and info['lnb']['v1_pointed']):
        # If a dual-polarization LNB is already pointed for Blocksat v1,
        # then we must use the polarization that the LNB was pointed to
        # originally, regardless of the satellite signal's polarization. In
        # v1, what mattered the most was the power supply voltage, which
        # determined the polarization of the dual polarization LNBs. If the
        # power supply provides voltage >= 18 (often the case), then the LNB
        # necessarily operates currently with horizontal polarization. Thus,
        # the same polarization must be configured in the S400.
        if (info['lnb']["v1_psu_voltage"] >= 16): # 16VDC is a common threshold
            pol = "H"
        else:
            pol = "V"
    else:
        if (info['sat']['pol'] == "H"):
            pol = "H"
        else:
            pol = "V"

    pol_label = "Horiz./L" if pol == "H" else "Vert./R"

    _item("Polarization: {}".format(pol_label))

    if (info['lnb']['universal'] and info['freqs']['dl'] > defs.ku_band_thresh):
        _item("Band (Tone): \"High/On\"")
    else:
        _item("Band (Tone): \"Low/Off\"")
    _item("Long Line Compensation: Disabled")
    _item("Apply")
    print()

    print("3. Verify that the S400 is locked to Blockstream Satellite's signal")
    _item("Check the \"RF 1 Lock\" indicator at the top of the page or the "
          "status LED in the S400's front panel. It should be green (locked) "
          "if your antenna is already pointed correctly. If not, you can work "
          "on the antenna pointing afterwards.")
    print()

    print("4. Go to Services > Tun1:\n")
    print("Scroll to \"Manage MPE PIDs\"")
    for pid in defs.pids:
        print("- Enter %d on \"New PID\" and click \"Add\"." %(pid))
    _item("Apply")
    print()

    print("** Optional configurations:")
    _item("If you prefer to use another IP address on LAN1 or LAN2, go to "
          "Interfaces > Data (LAN1) or Interfaces > M&C (LAN2) and "
          "configure the IP addresses. Note LAN 1 is the interface that "
          "will deliver the data packets received over satellite, whereas "
          "LAN2 is optional and exclusively for management.")
    print()

    util.prompt_for_enter()

    if (info['lnb']['pol'].lower() == "dual" and info['lnb']['v1_pointed'] and
        (pol != info['sat']['pol'])):
        util._print_sub_header("Notes")

        _item("The polarization that was suggested above assumes that you "
              "are going to use an LNB that is already pointed to {}, on a "
              "pre-existing SDR-based setup used for reception of the "
              "previous version of Blockstream Satellite (prior to the "
              "update to DVB-S2). It also assumes that you "
              "are not going to change the skew (polarization angle) of the "
              "LNB. The suggested polarization is exactly the same on which "
              "your LNB has been operating so far, i.e. {} polarization. "
              "However, note that your satellite signal has {} polarization. "
              "If you plan on re-pointing the LNB, please re-run the "
              "configuration helper (\"blocksat-cli cfg\") and "
              "answer that you are not reusing an already pointed LNB.".format(
                  info['sat']['name'],
                  "horizontal" if pol == "H" else "vertical",
                  "horizontal" if info['sat']['pol'] == "H" else "vertical"
              ))

        util.prompt_for_enter()

    util._print_sub_header("Host Configuration")

    _print("""
    In order to receive the traffic from the S400, you will need some networking
    configurations on your host. Such configurations are indicated and executed
    by running:
    """)

    print("\n```\nblocksat-cli standalone -i ifname\n```\n")
    print(textwrap.fill("where \'ifname\' should be replaced with the name "
                        "of the network interface that is connected to the "
                        "S400."))


def _print_usb_rx_instructions(info):
    """Print instructions for runnning with a Linux USB receiver
    """

    name = (info['setup']['vendor'] + " " + info['setup']['model']).strip()

    util._print_header(name)

    _print("""
    The {0} is a USB demodulator, which will receive data from satellite and
    will output data to the host over USB. The host, in turn, is responsible for
    configuring the demodulator using specific DVB-S2 tools. Hence, next, you
    need to prepare the host for driving the {0}.
    """.format(name))

    util._print_sub_header("Hardware Connections")

    print("The {} should be connected as follows:\n".format(name))

    print(("LNB ----> {0} (LNB Interface) -- "
           "{0} (USB Interface) ----> Host\n".format(name)))

    _item("Connect the LNB directly to \"LNB IN\" of the {} using a coaxial"
          " cable (an RG6 cable is recommended).".format(name))
    _item("Connect the {}'s USB interface to your computer.".format(name))

    util.prompt_for_enter()

    util._print_sub_header("Drivers")

    _print("""
    Before anything else, note that specific device drivers are required in order to
    use the {0}. Please, do note that driver installation can cause corruptions
    and, therefore, it is safer and **strongly recommended** to use a virtual
    machine for running the {0}. If you do so, please note that all commands
    recommended in the remainder of this page are supposed to be executed in the
    virtual machine.
    """.format(name))
    _print("""
    Next, install the drivers for the {0}. A helper script is available in the
    `util` directory from the root of this repository:
    """.format(name))

    _item("From the util/ folder, run:")
    print("""
    ./tbsdriver.sh
    """)
    print("Once the script completes the installation, reboot the virtual machine.")

    util.prompt_for_enter()

    util._print_sub_header("Host Requirements")

    print("Now, install all pre-requisites (in the virtual machine):")

    install_info = """
    On Ubuntu/Debian:

    sudo apt apt update
    sudo apt install python3 iproute2 iptables dvb-apps dvb-tools


    On Fedora

    sudo dnf update
    sudo dnf install python3 iproute iptables dvb-apps v4l-utils
    """
    print(install_info)

    _print("""
    If dvb-apps is not available on your distribution (for example on Fedora 31
    and 32), you can build it from source by running:
    """)

    build_info = """    git clone https://github.com/Blockstream/dvb-apps
    cd dvb-apps
    make
    sudo make install
    """

    print(build_info)

    util.prompt_for_enter()

    util._print_sub_header("Configure the Host")

    print("Run the following as root:")

    print("\n    sudo blocksat-cli usb config\n")

    _print("""
    This script will create network interfaces in order to handle the IP traffic
    received via the satellite link. It will define arbitrary IP addresses to the
    interfaces. To define a specific IP instead, use command-line argument `--ip`.
    """)

    _print(
        "NOTE: root privileges are required in order to configure firewall "
        "and reverse path (RP) filtering, as well as accessing the adapter "
        "at `/dev/dvb`. You will be prompted to accept or refuse the "
        "firewall and RP configurations.")

    util.prompt_for_enter()

    util._print_sub_header("Launch")

    print("Finally, launch the DVB-S2 interface by running:")

    print("\n    blocksat-cli usb launch\n")

    util.prompt_for_enter()

def _print_sdr_instructions(info):
    """Print instruction for configuration of an SDR setup
    """
    util._print_header("SDR Setup")

    util._print_sub_header("Connections")

    print("The SDR setup is connected as follows:\n")

    print("LNB ----> Power Supply ----> RTL-SDR ----> Host\n")

    _item("Connect the RTL-SDR USB dongle to your host PC.")
    _item("Connect the **non-powered** port of the power supply (labeled as "
          "\"Signal to IRD\") to the RTL-SDR using an SMA cable and an "
          "SMA-to-F adapter.")
    _item("Connect the **powered** port (labeled \"Signal to SWM\") of the "
          "power supply to the LNB using a coaxial cable (an RG6 cable is "
          "recommended).")

    print()
    _print("IMPORTANT: Do NOT connect the powered port of the power supply "
           "to the SDR interface. Permanent damage may occur to your SDR "
           "and/or your computer.")

    util.prompt_for_enter()

    util._print_sub_header("Software Requirements")

    print("The SDR-based setup relies on the applications listed below:\n")

    _item("leandvb: a software-based DVB-S2 demodulator.")
    _item("rtl_sdr: reads samples taken by the RTL-SDR and feeds them into leandvb.")
    _item("TSDuck: unpacks the output of leandvb and produces "
          "IP packets to be fed to Bitcoin Satellite.")
    _item("Gqrx: useful for spectrum visualization during antenna pointing.")

    print("\nTo install them, run:")
    print("""
    blocksat-cli deps install
    """)

    _print(
        """
        NOTE: This command assumes and has been tested with Ubuntu
        18.04 and Fedora 31. Please adapt if necessary in case you
        are using another Linux distribution or version.""")

    _print("If you prefer to build all software components manually, please "
           "refer to the SDR Guide at \"doc/sdr.md\" or:")
    print("https://github.com/Blockstream/satellite/blob/master/doc/sdr.md")

    util.prompt_for_enter()

    util._print_sub_header("Configuration")

    _print("Next, you can generate the configurations that are needed for gqrx "
           "by running:")

    print("    blocksat-cli gqrx-conf")

    util.prompt_for_enter()

    util._print_sub_header("Running")

    _print("You should now be ready to launch the SDR receiver. You can run it "
           "by executing:")

    print("    blocksat-cli sdr\n")
    print("Or, in GUI mode:\n")
    print("    blocksat-cli sdr --gui\n")

    _print("For further options, please refer to the SDR Guide "
           "at \"doc/sdr.md\" or:")
    print("https://github.com/Blockstream/satellite/blob/master/doc/sdr.md")

    util.prompt_for_enter()


def _print_freq_info(info):
    """Print summary of frequencies of interest"""
    sat     = info['sat']
    setup   = info['setup']
    lnb     = info['lnb']
    lo_freq = info['freqs']['lo']
    l_freq  = info['freqs']['l_band']

    util._print_header("Frequencies")

    print("| Downlink %2s band frequency                     | %8.2f MHz |" %(sat['band'], sat['dl_freq']))
    print("| Your LNB local oscillator (LO) frequency       | %8.2f MHz |" %(lo_freq))
    print("| L-band frequency to configure on your receiver | %7.2f MHz  |" %(l_freq))
    print()

    if (lnb['universal'] and (setup['type'] == defs.sdr_setup_type)):
        if (sat['dl_freq'] > defs.ku_band_thresh):
            print("NOTE regarding Universal LNB:\n")

            print(textwrap.fill(("The DL frequency of {} is in Ku high "
                                 "band (> {:.1f} MHz). Hence, you need to use "
                                 "the higher frequency LO ({:.1f} MHz) of your "
                                 "Universal LNB. This requires a 22 kHz tone "
                                 "to be sent to the LNB."
            ).format(sat['alias'], defs.ku_band_thresh, lo_freq)))
            print()
            print(textwrap.fill(("With a software-defined setup, you will "
                                 "need to place a 22 kHz tone generator "
                                 "inline between the LNB and the power "
                                 "inserter. Typically the tone generator "
                                 "uses power from the power inserter while "
                                 "delivering the tone directly to the "
                                 "LNB.")))

    util.prompt_for_enter()


def _print_lnb_info(info):
    """Print important waraning based on LNB choice"""
    lnb   = info['lnb']
    sat   = info['sat']
    setup = info['setup']

    if ((lnb['pol'] != "Dual") and (lnb['pol'] != sat['pol'])):
        util._print_header("LNB Information")
        lnb_pol = "Vertical" if lnb['pol'] == "V" else "Horizontal"
        logging.warning(textwrap.fill(
            "Your LNB has {} polarization and the signal from {} has the "
            "opposite polarization.".format(lnb_pol, sat['name'])))
        util.prompt_for_enter()

    if ((lnb['pol'] == "Dual") and (setup['type'] == defs.sdr_setup_type)):
        util._print_header("LNB Information")
        logging.warning(textwrap.fill(
            "Your LNB has dual polarization. Check the voltage of your power "
            "supply in order to discover the polarization on which your LNB "
            "will operate."))
        util.prompt_for_enter()


def _print_next_steps():
    util._print_header("Next Steps")
    _print("""
    At this point, if your dish is already correctly pointed, you should be able to
    start receiving data in Bitcoin Satellite.
    """)

    print("You can generate a bitcoin.conf configuration file for Bitcoin Satellite using:")
    print("\n    blocksat-cli btc\n")

    print("For further information, refer to \"doc/bitcoin.md\" or:\n")
    print("https://github.com/Blockstream/satellite/blob/master/doc/bitcoin.md\n")

    _print("""If your antenna is not pointed yet, please follow the
    antenna alignment guide available at \"doc/antenna-pointing.md\" or:""")
    print("https://github.com/Blockstream/satellite/blob/master/doc/antenna-pointing.md\n")


def subparser(subparsers):
    """Argument parser of instructions command"""
    p = subparsers.add_parser('instructions',
                              description="Instructions for Blocksat Rx setup",
                              help='Read instructions for the receiver setup',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.set_defaults(func=show)
    return p


def show(args):
    """Show instructions"""
    info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (info is None):
        return

    _print_freq_info(info)
    _print_lnb_info(info)

    if (info['setup']['type'] == defs.standalone_setup_type):
        _print_s400_instructions(info)
    elif (info['setup']['type'] == defs.sdr_setup_type):
        _print_sdr_instructions(info)
    elif (info['setup']['type'] == defs.linux_usb_setup_type):
        _print_usb_rx_instructions(info)

    _print_next_steps()



"""User setup configuration"""
import os, json, logging
from argparse import ArgumentDefaultsHelpFormatter, Namespace
from pprint import pprint, pformat
from . import util, defs, instructions
import textwrap
from decimal import Decimal, getcontext


def _cfg_satellite():
    """Configure satellite covering the user"""

    util._print_header("Satellite")

    help_msg = "Not sure? Check the coverage map at:\n" \
               "https://blockstream.com/satellite/#satellite_network-coverage"

    question = "Please, inform which satellite below covers your location:"
    sat = util._ask_multiple_choice(defs.satellites,
                                    question,
                                    "Satellite",
                                    lambda sat : '{} ({})'.format(sat['name'],
                                                                  sat['alias']),
                                    help_msg)
    return sat


def _cfg_rx_setup():
    """Configure Rx setup - which receiver user is using """

    util._print_header("Receiver Setup")

    # Demodulator
    question = "Please, inform your DVB-S2 receiver setup from the list below:"
    setup = util._ask_multiple_choice(
        defs.demods,
        question,
        "Setup",
        lambda x : '{} receiver, using {} demodulator'.format(
            x['type'], (x['vendor'] + " " + x['model']).strip()))

    # Network interface connected to the standalone demodulator
    if (setup['type'] == defs.standalone_setup_type):
        try:
            devices = os.listdir('/sys/class/net/')
        except FileNotFoundError:
            devices = None
            pass

        question = "Which network interface is connected to the {}?".format(
            setup['model'])
        if (devices is not None):
            netdev = util._ask_multiple_choice(devices,
                                               question,
                                               "Interface",
                                               lambda x : '{}'.format(x))
        else:
            netdev = input(question + " ")

        setup['netdev'] = netdev.strip()

    # Antenna
    question = "Please, inform the type of your satellite dish (antenna):"
    size = util._ask_multiple_choice(
        defs.antennas, question, "Size",
        lambda x : 'Satellite Dish ({})'.format(x['label'])
        if x['type'] == 'dish'
        else '{} Flat-Panel Antenna'.format(x['label']),
        none_option = True,
        none_str = "Other")

    if (size is None):
        resp = util.typed_input("Enter size in cm",
                                "Please enter an integer number in cm",
                                in_type=int)
        size = { 'label' : "custom", 'size' : resp }

    setup['antenna'] = size

    return setup


def _cfg_custom_lnb(sat):
    """Configure custom LNB based on user-entered specs

    Args:
        sat : user's satellite info

    """

    print("\nPlease inform the specifications of your LNB:")

    bands           = ["C", "Ku"]
    question        = "Frequency band:"
    custom_lnb_band = util._ask_multiple_choice(bands, question, "Band",
                                                lambda x : '{}'.format(x))

    if (sat['band'].lower() != custom_lnb_band.lower()):
        logging.error(
            "You must use a %s band LNB in order to receive from %s" %(
                sat['band'], sat['name']))
        exit(1)


    if (custom_lnb_band == "Ku"):
        custom_lnb_universal = util._ask_yes_or_no("Is it a Universal Ku band LNB?")

        if (custom_lnb_universal):
            print(textwrap.fill(
                "A Universal Ku band LNB has two LO (local oscillator) " + \
                " frequencies. Typically the two frequencies are 9750 MHz " +
                "and 10600 MHz."))
            if (util._ask_yes_or_no("Does your LNB have LO frequencies 9750 MHz and 10600 MHz?")):
                custom_lnb_lo_freq = [9750.0, 10600]
            else:
                custom_lnb_lo_freq = []
                while (len(custom_lnb_lo_freq) != 2):
                    try:
                        resp = input("Inform the two LO frequencies in MHz, "
                                     "separated by comma: ")
                        custom_lnb_lo_freq = [float(x) for x in resp.split(",")]
                    except ValueError:
                        continue

        else:
            # Non-universal Ku-band LNB
            custom_lnb_lo_freq = util.typed_input("LNB LO frequency in MHz",
                                                  in_type=float)
    else:
        # C-band LNB
        custom_lnb_universal = False
        custom_lnb_lo_freq = util.typed_input("LNB LO frequency in MHz",
                                              in_type=float)

    # Polarization
    question = "Choose the LNB polarization:"
    options = [
        {
            'id' : "Dual",
            'text' : "Dual polarization (horizontal and vertical)"
        },
        {
            'id' : "H",
            'text' : "Horizontal"
        },
        {
            'id' : "V",
            'text' : "Vertical"
        }]
    pol = util._ask_multiple_choice(options, question,
                                    "Polarization",
                                    lambda x : '{}'.format(x['text']))

    return {
        'vendor'    : "",
        'model'     : "",
        "lo_freq"   : custom_lnb_lo_freq,
        'universal' : custom_lnb_universal,
        'band'      : custom_lnb_band,
        'pol'       : pol['id']
    }


def _cfg_lnb(sat, setup):
    """Configure LNB - either from preset or from custom specs

    Args:
        sat   : user's satellite info
        setup : user's setup info

    """

    if (setup['antenna']['type'] == 'flat'):
        for lnb in defs.lnbs:
            if lnb['vendor'] == 'Selfsat':
                lnb["v1_pointed"] = False
                return lnb

    util._print_header("LNB")

    question = "Please, inform your LNB model:"
    lnb = util._ask_multiple_choice(
        defs.lnbs, question, "LNB",
        lambda x : "{} {} {}".format(
            x['vendor'], x['model'],
            "(Universal Ku band LNBF)" if x['universal'] else ""),
        none_option = True,
        none_str = "Other")

    if (lnb is None):
        lnb = _cfg_custom_lnb(sat)

    if (sat['band'].lower() != lnb['band'].lower()):
        logging.error("The LNB you chose cannot operate " + \
                      "in %s band (band of satellite %s)" %(sat['band'],
                                                            sat['alias']))
        exit(1)

    # For dual polarization LNBs, we must know whether it was pointed before for
    # blocksat v1 in order to define the polarization on channels.conf
    if (lnb['pol'].lower() == "dual"):
        prev_setup = util._ask_yes_or_no(
            "Are you reusing an LNB that is already pointed and that was used "
            "for the previous version of Blockstream Satellite (before the "
            "upgrade to DVB-S2)?",
        help_msg="NOTE: this information is helpful to determine the "
            "polarization required for the LNB.")

        if (prev_setup):
            question = ("In this setup, did you use one of the LNB power "
                        "inserters below?")
            psu = util._ask_multiple_choice(
                defs.psus, question, "Power inserter",
                lambda x : "{}".format(x['model']),
                none_option = True,
                none_str = "No - another model")
            if (psu is None):
                voltage = util.typed_input("What is the voltage supplied to the "
                                           "LNB by your power inserter?")
            else:
                voltage = psu["voltage"]

            lnb["v1_pointed"]     = True
            lnb["v1_psu_voltage"] = voltage
        else:
            lnb["v1_pointed"] = False

    return lnb


def _cfg_frequencies(sat, lnb):
    """Print summary of frequencies

    Inform the downlink RF frequency, the LNB LO frequency and the L-band
    frequency to be configured in the receiver.

    Args:
        sat   : user's satellite info
        lnb   : user's LNB info

    """
    getcontext().prec = 8

    if (sat['band'].lower() == "ku"):
        if (lnb['universal']):
            assert(isinstance(lnb['lo_freq'], list)), \
                "A Universal LNB must have a list with two LO frequencies"
            assert(len(lnb['lo_freq']) == 2), \
                "A Universal LNB must have two LO frequencies"

            if (sat['dl_freq'] > defs.ku_band_thresh):
                lo_freq = lnb['lo_freq'][1]
            else:
                lo_freq = lnb['lo_freq'][0]
        else:
            lo_freq = lnb['lo_freq']

        if_freq = float(Decimal(sat['dl_freq']) - Decimal(lo_freq))

    elif (sat['band'].lower() == "c"):
        lo_freq = lnb['lo_freq']
        if_freq = float(Decimal(lo_freq) - Decimal(sat['dl_freq']))
    else:
        raise ValueError("Unknown satellite band")

    return {
        'dl'     : sat['dl_freq'],
        'lo'     : lo_freq,
        'l_band' : if_freq
    }


def _cfg_chan_conf(info, chan_file):
    """Generate the channels.conf file"""

    util._print_header("Channel Configurations for Linux USB Rx")

    print(textwrap.fill("This step will generate the channel configuration "
                        "file that is required when launching the USB "
                        "receiver in Linux.") + "\n")

    if (os.path.isfile(chan_file)):
        print("Found previous %s file:" %(chan_file))

        if (not util._ask_yes_or_no("Remove and regenerate file?")):
            print("Configuration aborted.")
            return
        else:
            os.remove(chan_file)

    with open(chan_file, 'w') as f:
        f.write('[blocksat-ch]\n')
        f.write('\tDELIVERY_SYSTEM = DVBS2\n')
        f.write('\tFREQUENCY = %u\n' %(int(info['sat']['dl_freq']*1000)))
        if (info['lnb']['pol'].lower() == "dual" and info['lnb']['v1_pointed']):
            # If a dual-polarization LNB is already pointed for Blocksat v1,
            # then we must use the polarization that the LNB was pointed to
            # originally, regardless of the satellite signal's polarization. In
            # v1, what mattered the most was the power supply voltage, which
            # determined the polarization of the dual polarization LNBs. If the
            # power supply provides voltage >= 18 (often the case), then the LNB
            # necessarily operates currently with horizontal polarization. Thus,
            # on channels.conf we must use the same polarization in order for
            # the DVB adapter to supply the 18VDC voltage.
            if (info['lnb']["v1_psu_voltage"] >= 16): # 16VDC threshold
                f.write('\tPOLARIZATION = HORIZONTAL\n')
            else:
                f.write('\tPOLARIZATION = VERTICAL\n')
        else:
            if (info['sat']['pol'] == 'V'):
                f.write('\tPOLARIZATION = VERTICAL\n')
            else:
                f.write('\tPOLARIZATION = HORIZONTAL\n')
        f.write('\tSYMBOL_RATE = {}\n'.format(
            defs.sym_rate[info['sat']['alias']]))
        f.write('\tINVERSION = AUTO\n')
        f.write('\tMODULATION = QPSK\n')
        f.write('\tVIDEO_PID = 32+33\n')

    print("File \"%s\" saved." %(chan_file))

    with open(chan_file, 'r') as f:
        logging.debug(f.read())


def _read_cfg_file(cfg_file):
    """Read configuration file"""

    if (os.path.isfile(cfg_file)):
        with open(cfg_file) as fd:
            info = json.load(fd)
        return info


def _rst_cfg_file(cfg_file):
    """Reset a previous configuration file in case it exists"""
    info = _read_cfg_file(cfg_file)

    if (info is not None):
        print("Found previous configuration:")
        pprint(info, width=80, compact=False)
        if (util._ask_yes_or_no("Reset?")):
            os.remove(cfg_file)
        else:
            print("Configuration aborted.")
            return False
    return True


def read_cfg_file(basename, directory):
    """Read configuration file

    If not available, run configuration helper.

    """
    cfg_file = os.path.join(directory, os.path.basename(basename))
    info = _read_cfg_file(cfg_file)

    while (info is None):
        print("Missing {} configuration file".format(cfg_file))
        if (util._ask_yes_or_no("Run configuration helper now?")):
            configure(Namespace(cfg_dir=directory,
                                cfg_file=basename))
        else:
            print("Abort")
            return

        info = _read_cfg_file(cfg_file)

    return info


def subparser(subparsers):
    """Argument parser of config command"""
    p = subparsers.add_parser('config', aliases=['cfg'],
                              description="Configure Blocksat Rx setup",
                              help='Define receiver and Bitcoin Satellite \
                              configurations',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-c', '--chan-conf',
                   default="channels.conf",
                   help='Channel configurations file')
    p.set_defaults(func=configure)
    return p


def configure(args):
    """Configure Blocksat Receiver setup

    """
    cfg_file = os.path.join(args.cfg_dir, os.path.basename(args.cfg_file))
    rst_ok   = _rst_cfg_file(cfg_file)
    if (not rst_ok):
        return

    user_sat   = _cfg_satellite()
    user_setup = _cfg_rx_setup()
    user_lnb   = _cfg_lnb(user_sat, user_setup)
    user_freqs = _cfg_frequencies(user_sat, user_lnb)

    user_info = {
        'sat'   : user_sat,
        'setup' : user_setup,
        'lnb'   : user_lnb,
        'freqs' : user_freqs
    }

    logging.debug(pformat(user_info))

    if not os.path.exists(args.cfg_dir):
        os.makedirs(args.cfg_dir)

    with open(cfg_file, 'w') as fd:
        json.dump(user_info, fd)

    util._print_header("JSON configuration file")
    print("Saved configurations on %s" %(cfg_file))

    if (user_setup['type'] == defs.linux_usb_setup_type):
        if 'chan_conf' in args:
            chan_file = os.path.join(args.cfg_dir, args.chan_conf)
        else:
            chan_file = os.path.join(args.cfg_dir, "channels.conf")
        _cfg_chan_conf(user_info, chan_file)

    util._print_header("Next Steps")

    print(textwrap.fill(
        "Please check setup instructions by running:"))
    print("""
    blocksat-cli instructions
    """)





#!/usr/bin/env python3
import logging, os
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from . import defs, config, util, instructions, gqrx, bitcoin, sdr, rp, \
    firewall, standalone, usb, dependencies
from os import environ
import platform


__version__ = "0.1.5"


def main():
    """Main - parse command-line arguments and call subcommands
    """
    assert(platform.system() == 'Linux'), \
        "blocksat-cli currently only supports Linux"
    sudo_user       = environ.get('SUDO_USER')
    user            = sudo_user if sudo_user is not None else ""
    home            = os.path.expanduser("~" + user)
    default_cfg_dir = os.path.join(home, ".blocksat")

    parser = ArgumentParser(prog="blocksat-cli",
                            description="Blockstream Satellite Command-Line Interface",
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-d', '--debug', action='store_true',
                        help='Set debug mode')
    parser.add_argument('--cfg-file',
                        default="config.json",
                        help="Configuration file name")
    parser.add_argument('--cfg-dir',
                        default=default_cfg_dir,
                        help="Directory to use for configuration files")
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s {}'.format(__version__))

    subparsers = parser.add_subparsers(title='subcommands',
                                       help='Target sub-command')

    config.subparser(subparsers)
    instructions.subparser(subparsers)
    dependencies.subparser(subparsers)
    usb.subparser(subparsers)
    standalone.subparser(subparsers)
    rp.subparser(subparsers)
    firewall.subparser(subparsers)
    gqrx.subparser(subparsers)
    bitcoin.subparser(subparsers)
    sdr.subparser(subparsers)

    args = parser.parse_args()
    log_format = '%(levelname)s: %(message)s'
    if (args.debug):
        logging.basicConfig(level=logging.DEBUG, format=log_format)
        logging.debug('[Debug Mode]')
    else:
        logging.basicConfig(level=logging.INFO, format=log_format)

    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()



"""Configure Firewall Rules"""
import subprocess, logging, os
from argparse import ArgumentDefaultsHelpFormatter
from . import util, defs


def _get_iptables_rules(net_if):
    """Get iptables rules that are specifically applied to a target interface

    Args:
        net_if : network interface name

    Returns:
        list of dictionaries with information of the individual matched rules

    """



    # Get rules
    cmd = util.root_cmd(["iptables", "-L", "-v", "--line-numbers"])
    res = subprocess.check_output(cmd)

    # Parse
    header1 = ""
    header2 = ""
    rules   = list()
    for line in res.splitlines():
        if ("Chain INPUT" in line.decode()):
            header1 = line.decode()

        if ("destination" in line.decode()):
            header2 = line.decode()

        if (net_if in line.decode()):
            rules.append({
                'rule' : line.decode().split(),
                'header1' : header1,
                'header2' : header2
            })

    return rules


def _is_iptables_igmp_rule_set(net_if, cmd):
    """Check if an iptables rule for IGMP is already configured

    Args:
        net_if : network interface name
        cmd    : list with iptables command

    Returns:
        True if rule is already set, False otherwise.

    """
    offset = 1 if cmd[0] == "sudo" else 0

    for rule in _get_iptables_rules(net_if):
        if (rule['rule'][3] == "ACCEPT" and
            rule['rule'][6] == cmd[6 + offset] and
            rule['rule'][4] == "igmp"):
            print("\nFirewall rule for IGMP already configured\n")
            print(rule['header1'])
            print(rule['header2'])
            print(" ".join(rule['rule']))
            print("\nSkipping...")
            return True

    return False


def _is_iptables_udp_rule_set(net_if, cmd):
    """Check if an iptables rule for UDP is already configured

    Args:
        net_if : network interface name
        cmd    : list with iptables command

    Returns:
        True if rule is already set, False otherwise.

    """
    offset = 1 if cmd[0] == "sudo" else 0
    for rule in _get_iptables_rules(net_if):
        if (rule['rule'][3] == "ACCEPT" and
            rule['rule'][6] == cmd[6 + offset] and
            (rule['rule'][4] == "udp" and
             rule['rule'][12] == cmd[10 + offset])):
            print("\nFirewall rule already configured\n")
            print(rule['header1'])
            print(rule['header2'])
            print(" ".join(rule['rule']))
            print("\nSkipping...")
            return True

    return False


def _add_iptables_rule(net_if, cmd):
    """Add iptables rule

    Args:
        net_if : network interface name
        cmd    : list with iptables command

    """
    offset = 1 if cmd[0] == "sudo" else 0

    # Set up iptables rules
    logging.debug("> " + " ".join(cmd))
    subprocess.check_output(cmd)

    for rule in _get_iptables_rules(net_if):
        print_rule = False

        if (rule['rule'][3] == "ACCEPT" and
            rule['rule'][6] == cmd[6 + offset] and
            rule['rule'][4] == cmd[4 + offset]):
            if (cmd[4 + offset] == "igmp"):
                print_rule = True
            elif (cmd[4 + offset] == "udp" and
                  rule['rule'][12] == cmd[10 + offset]):
                print_rule = True

            if (print_rule):
                print("Added iptables rule:\n")
                print(rule['header1'])
                print(rule['header2'])
                print(" ".join(rule['rule']) + "\n")


def _configure_firewall(net_if, ports, igmp=False, prompt=True):
    """Configure firewallrules to accept blocksat traffic via DVB interface

    Args:
        net_if : DVB network interface name
        ports  : ports used for blocks traffic and API traffic
        igmp   : Whether or not to configure rule to accept IGMP queries

    """

    not_root = (os.geteuid() != 0)

    cmd = util.root_cmd([
        "iptables",
        "-I", "INPUT",
        "-p", "udp",
        "-i", net_if,
        "--match", "multiport",
        "--dports", ",".join(ports),
        "-j", "ACCEPT",
    ])

    if (not_root):
        print(" ".join(cmd))
    elif (not _is_iptables_udp_rule_set(net_if, cmd)):
        print("- Configure firewall rule to accept Blocksat traffic arriving " +
              "at interface %s\ntowards UDP ports %s." %(net_if, ",".join(ports)))
        if ((not prompt) or
            util._ask_yes_or_no("Add corresponding ACCEPT firewall rule?")):
            _add_iptables_rule(net_if, cmd)
        else:
            print("\nFirewall configuration cancelled")


    # We're done, unless we also need to configure an IGMP rule
    if (not igmp):
        return

    # IGMP rule supports standalone DVB demodulators. The host in this case will
    # need to periodically send IGMP membership reports in order for upstream
    # switches between itself and the DVB demodulator to continue delivering the
    # multicast-addressed traffic. This overcomes the scenario where group
    # membership timeouts are implemented by the intermediate switches.
    cmd = util.root_cmd([
        "iptables",
        "-I", "INPUT",
        "-p", "igmp",
        "-i", net_if,
        "-j", "ACCEPT",
    ])

    if (not_root):
        print(" ".join(cmd))
    elif (not _is_iptables_igmp_rule_set(net_if, cmd)):
        print("Configure also a firewall rule to accept IGMP queries. This is " +
              "necessary when using a standalone DVB demodulator.")
        if ((not prompt) or
            util._ask_yes_or_no("Add corresponding ACCEPT rule on firewall?")):
            _add_iptables_rule(net_if, cmd)
        else:
            print("\nIGMP firewall rule cancelled")


def configure(net_ifs, ports, igmp=False, prompt=True):
    """Configure firewallrules to accept blocksat traffic via DVB interface

    Args:
        net_ifs : List of DVB network interface names
        ports   : ports used for blocks traffic and API traffic
        igmp    : Whether or not to configure rule to accept IGMP queries

    """
    assert(isinstance(net_ifs, list))
    print("\n------------------------------- Firewall Rules " +
          "---------------------------------")

    if (os.geteuid() != 0):
        util.fill_print("Please run blocksat-cli as root or run the following \
        commands on your own:\n")

    for i, net_if in enumerate(net_ifs):
        _configure_firewall(net_if, ports, igmp, prompt)

        if (i < len(net_ifs) - 1):
            print("")


def subparser(subparsers):
    """Parser for firewall command"""
    p = subparsers.add_parser('firewall',
                              description="Set firewall rules",
                              help='Set firewall rules',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-i', '--interface', required = True,
                   help='Network interface')
    p.add_argument('--standalone', default=False,
                   action='store_true',
                   help='Configure for standalone DVB-S2 demodulator')
    p.set_defaults(func=firewall_subcommand)
    return p


def firewall_subcommand(args):
    """Call function that sets firewall rules

    Handles the firewall subcommand

    """
    configure([args.interface], defs.src_ports, igmp=args.standalone)





"""Standalone Demodulator"""
from argparse import ArgumentDefaultsHelpFormatter
from . import rp, firewall, defs, config


def subparser(subparsers):
    p = subparsers.add_parser('standalone',
                              description="Configure host to receive \
                              data from standalone DVB-S2 demodulator",
                              help='Configure host to receive from \
                              standalone DVB-S2 demodulator',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-i', '--interface',
                    default=None,
                    help='Network interface connected to the standalone \
                    demodulator')
    p.add_argument('-y', '--yes', default=False, action='store_true',
                   help="Default to answering Yes to configuration prompts")
    p.set_defaults(func=cfg_standalone)
    return p

def cfg_standalone(args):
    """Configurations for standalone DVB demodulator
    """
    # User info
    user_info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if 'netdev' not in user_info['setup']:
        assert(args.interface is not None), \
            ("Please specify the network interface through option "
             "\"-i/--interface\"")

    interface = args.interface if (args.interface is not None) else \
                user_info['setup']['netdev']

    rp.set_filters([interface], prompt=(not args.yes))
    firewall.configure([interface], defs.src_ports, igmp=True,
                       prompt=(not args.yes))





import re, sys
from setuptools import setup


if sys.version_info[0] < 3:
    raise SystemExit("Error: blocksat-cli requires Python 3")
    sys.exit(1)

version = re.search(
    '^__version__\s*=\s*"(.*)"',
    open('blocksatcli/main.py').read(),
    re.M
).group(1)

long_description = """# Blockstream Satellite CLI

A command-line interface for configuring, running and monitoring a Blockstream
Satellite receiver setup.

"""

setup(
    name = "blocksat-cli",
    packages = ["blocksatcli"],
    entry_points = {
        "console_scripts": ['blocksat-cli = blocksatcli.main:main']
    },
    version = version,
    description = "Blockstream Satellite CLI",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author = "Blockstream Corp",
    author_email = "satellite@blockstream.com",
    url = "https://github.com/Blockstream/satellite",
    classifiers=[
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    python_requires='>=3'
)



"""Bitcoin .conf generator"""
import os
from argparse import ArgumentDefaultsHelpFormatter
from . import config, defs, util
import textwrap
import argparse


def _udpmulticast(dev, src_addr, dst_addr=defs.btc_dst_addr, trusted="1",
                  label=""):
    """Return the udpmulticast configuration line for bitcoin.conf"""
    return "udpmulticast=" + dev + "," + dst_addr + "," + src_addr + "," + trusted + "," + label


def subparser(subparsers):
    """Argument parser of bitcoin-conf command"""
    p = subparsers.add_parser('bitcoin-conf', aliases=['btc'],
                              description="Generate Bitcoin configuration file",
                              help='Generate Bitcoin configuration file',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    group = p.add_mutually_exclusive_group()
    group.add_argument('-d', '--datadir', default=None,
                       help='Path to the data directory where the generated '
                       'bitcoin.conf will be saved')
    group.add_argument('--stdout', action='store_true', default=False,
                       help='Print bitcoin.conf configurations and don\'t save')
    p.set_defaults(func=configure)

    return subparser


def configure(args):
    """Generate bitcoin.conf configuration"""
    info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (info is None):
        return

    util._print_header("Bitcoin Conf Generator")

    if args.datadir is None:
        home = os.path.expanduser("~")
        path = os.path.join(home, ".bitcoin")
    else:
        path = args.datadir

    conf_file = "bitcoin.conf"
    abs_path  = os.path.join(path, conf_file)

    cfg = ("debug=udpnet\n"
           "debug=udpmulticast\n"
           "udpmulticastloginterval=10\n")

    if (info['setup']['type'] == defs.sdr_setup_type):
        cfg += _udpmulticast(dev="lo",
                             src_addr="127.0.0.1",
                             label="blocksat-sdr") + "\n"
    elif (info['setup']['type'] == defs.linux_usb_setup_type):
        cfg += _udpmulticast(dev="dvb0_0",
                             src_addr=info['sat']['ip'],
                             label="blocksat-tbs-lowspeed") + "\n"
        cfg += _udpmulticast(dev="dvb0_1",
                             src_addr=info['sat']['ip'],
                             label="blocksat-tbs-highspeed") + "\n"
    elif (info['setup']['type'] == defs.standalone_setup_type):
        cfg += _udpmulticast(dev=info['setup']['netdev'],
                             src_addr=info['sat']['ip'],
                             label="blocksat-s400") + "\n"
    else:
        raise ValueError("Unknown setup type")

    if (args.stdout):
        print(cfg)
        return

    print("Save {} at {}/".format(conf_file, path))

    if (not util._ask_yes_or_no("Proceed?")):
        print("Aborted")
        return

    if not os.path.exists(path):
        os.makedirs(path)

    if os.path.exists(abs_path):
        if (not util._ask_yes_or_no("File already exists. Overwrite?")):
            print("Aborted")
            return

    with open(abs_path, "w") as file:
        file.write(cfg)

    print("Saved")

    if (info['setup']['type'] == defs.linux_usb_setup_type):
        print("\n" + textwrap.fill(
            ("NOTE: {} was configured assuming the DVB-S2 net interfaces will "
             "be named dvb0_0 and dvb0_1. You can check if this is the case "
             "after launching the system. If it isn't, please update {} "
             "accordingly.").format(conf_file, conf_file)))
    elif (info['setup']['type'] == defs.standalone_setup_type):
        print("\n" + textwrap.fill(
            ("NOTE: {} was configured assuming the S400 will be connected to "
             "interface {}. If this isn't the case anymore, please update {} "
             "accordingly.").format(
                 conf_file, info['setup']['netdev'], conf_file)))




"""Generate gqrx configurations"""
import os
from argparse import ArgumentDefaultsHelpFormatter
from . import config, util
import textwrap
import argparse


def subparser(subparsers):
    p = subparsers.add_parser('gqrx-conf', aliases=['gqrx'],
                              description="Generate gqrx configurations",
                              help='Generate gqrx configurations',
                              formatter_class=ArgumentDefaultsHelpFormatter)
    p.add_argument('-p', '--path', default=None,
                   help='Path where gqrx configuration file should be saved')

    p.set_defaults(func=configure)

    return subparser


def configure(args):
    """Configure GQRX"""
    info = config.read_cfg_file(args.cfg_file, args.cfg_dir)

    if (info is None):
        return

    util._print_header("Gqrx Conf Generator")

    if args.path is None:
        home = os.path.expanduser("~")
        path = os.path.join(home, ".config", "gqrx")
    else:
        path = args.path

    conf_file = "default.conf"
    abs_path  = os.path.join(path, conf_file)

    cfg = """
    [General]
    configversion=2

    [fft]
    averaging=80
    db_ranges_locked=true
    pandapter_min_db=-90
    waterfall_min_db=-90

    [input]
    bandwidth=1000000
    device="rtl=0"
    frequency={}
    lnb_lo={}
    sample_rate=2400000

    [receiver]
    demod=0
    """.format(
        int(info['freqs']['dl']*1e6),
        int(info['freqs']['lo']*1e6),
    )

    print("Save {} at {}/".format(conf_file, path))

    if (not util._ask_yes_or_no("Proceed?")):
        print("Aborted")
        return

    if not os.path.exists(path):
        os.makedirs(path)

    if os.path.exists(abs_path):
        if (not util._ask_yes_or_no("File already exists. Overwrite?")):
            print("Aborted")
            return

    with open(abs_path, "w") as file:
        file.write(textwrap.dedent(cfg))

    print("Saved")




"""Configure IP Address of DVB Interfaces"""
from ipaddress import IPv4Interface
import os, subprocess, logging
from . import util
import textwrap
from shutil import which
logger = logging.getLogger(__name__)


def _check_debian_net_interfaces_d(is_root):
    """Check if /etc/network/interfaces.d/ is included as source-directory

    If the distribution doesn't have the `/etc/network/interface` directory,
    just return.

    """
    if_file  = "/etc/network/interfaces"

    if (not os.path.exists(if_file)):
        return

    if_dir   = "/etc/network/interfaces.d/"
    src_line = "source /etc/network/interfaces.d/*"

    if (not is_root):
        print(textwrap.fill("Make sure directory "
                            "\"{}\" "
                            "is considered as source for network "
                            "interface configurations. "
                            "Check if your \"{}\" file "
                            "contains the following line:".format(
                                if_dir, if_file
                            )))
        print("\n{}\n".format(src_line))
        print("If not, add it.\n")
        return

    with open(if_file) as fd:
        current_cfg = fd.read()

    src_included = False
    for line in current_cfg.splitlines():
        if src_line in line:
            src_included = True
            break

    if (src_included):
        return

    with open(if_file, 'a') as fd:
        fd.write("\n" + src_line)


def _add_to_netplan(ifname, addr_with_prefix, is_root):
    """Create configuration file at /etc/netplan/"""
    assert("/" in addr_with_prefix)
    cfg_dir = "/etc/netplan/"

    cfg = ("network:\n"
           "  version: 2\n"
           "  renderer: networkd\n"
           "  ethernets:\n"
           "    {0}:\n"
           "      dhcp4: no\n"
           "      optional: true\n"
           "      addresses: [{1}]\n").format(
               ifname, addr_with_prefix)

    fname  = "blocksat-" + ifname + ".yaml"
    path   = os.path.join(cfg_dir, fname)

    if (not is_root):
        print(textwrap.fill("Create a file named {} at {} and add the "
                            "following to it:".format(fname, cfg_dir)))
        print("\n" + cfg + "\n")
        return

    with open(path, 'w') as fd:
        fd.write(cfg)

    print("Created configuration file {} in {}".format(fname, cfg_dir))


def _add_to_interfaces_d(ifname, addr, netmask, is_root):
    """Create configuration file at /etc/network/interfaces.d/"""
    if_dir = "/etc/network/interfaces.d/"
    cfg = ("iface {0} inet static\n"
           "    address {1}\n"
           "    netmask {2}\n").format(ifname, addr, netmask)
    fname  = ifname + ".conf"
    path   = os.path.join(if_dir, fname)

    if (not is_root):
        print(textwrap.fill("Create a file named {} at {} and add the "
                            "following to it:".format(fname, if_dir)))
        print("\n" + cfg + "\n")
        return

    with open(path, 'w') as fd:
        fd.write(cfg)

    print("Created configuration file {} in {}".format(fname, if_dir))


def _add_to_sysconfig_net_scripts(ifname, addr, netmask, is_root):
    """Create configuration file at /etc/sysconfig/network-scripts/"""
    cfg_dir = "/etc/sysconfig/network-scripts/"
    cfg = ("NM_CONTROLLED=\"yes\"\n"
           "DEVICE=\"{0}\"\n"
           "BOOTPROTO=none\n"
           "ONBOOT=\"no\"\n"
           "IPADDR={1}\n"
           "NETMASK={2}\n").format(ifname, addr, netmask)

    fname  = "ifcfg-" + ifname
    path   = os.path.join(cfg_dir, fname)

    if (not is_root):
        print(textwrap.fill("Create a file named {} at {} and add the "
                            "following to it:".format(fname, cfg_dir)))
        print("\n" + cfg + "\n")
        return

    with open(path, 'w') as fd:
        fd.write(cfg)

    print("Created configuration file {} in {}".format(fname, cfg_dir))


def _set_static_iface_ip(ifname, ipv4_if, is_root):
    """Set static IP address for target network interface

    Args:
        ifname  : Network device name
        ipv4_if : IPv4Interface object with the target interface address
        is_root : (bool) whether running as root

    """
    isinstance(ipv4_if, IPv4Interface)
    addr                = str(ipv4_if.ip)
    addr_with_prefix    = ipv4_if.with_prefixlen
    netmask             = str(ipv4_if.netmask)

    if (which("netplan") is not None):
        _add_to_netplan(ifname, addr_with_prefix, is_root)
    elif (os.path.exists("/etc/network/interfaces.d/")):
        _add_to_interfaces_d(ifname, addr, netmask, is_root)
    elif (os.path.exists("/etc/sysconfig/network-scripts")):
        _add_to_sysconfig_net_scripts(ifname, addr, netmask, is_root)
    else:
        raise ValueError("Could not set a static IP address on interface "
                         "{}".format(ifname))


def _check_ip(net_if, ip_addr):
    """Check if interface has IP and if it matches target IP

    Args:
        net_if  : DVB network interface name
        ip_addr : Target IP address for the DVB interface slash subnet mask

    Returns:
        (Bool, Bool) Tuple of booleans. The first indicates whether interface
        already has an IP. The second indicates whether the interface IP (if
        existing) matches with respect to a target IP.

    """
    try:
        res = subprocess.check_output(["ip", "addr", "show", "dev", net_if],
                                      stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError as e:
        return False, False

    has_ip = False
    ip_ok  = False
    for line in res.splitlines():
        if "inet" in line.decode() and "inet6" not in line.decode():
            has_ip    = True
            # Check if IP matches target
            inet_info = line.decode().split()
            inet_if   = IPv4Interface(inet_info[1])
            target_if = IPv4Interface(ip_addr)
            ip_ok     = (inet_if == target_if)
            break

    return has_ip, ip_ok


def _set_ip(net_if, ip_addr, verbose):
    """Set the IP of the DVB network interface

    Args:
        net_if    : DVB network interface name
        ip_addr   : Target IP address for the DVB interface slash subnet mask
        verbose   : Controls verbosity

    """
    is_root       = (os.geteuid() == 0)
    inet_if       = IPv4Interface(ip_addr)

    # Flush previous IP
    if (not is_root):
        print("Flush any IP address from {}:\n".format(net_if))
    res = util.run_or_print_root_cmd(["ip", "address", "flush", "dev", net_if],
                                     logger)

    # Configure new static IP
    _set_static_iface_ip(net_if, inet_if, is_root)


def set_ips(net_ifs, ip_addrs, verbose=True):
    """Set IPs of one or multiple DVB network interface(s

    Args:
        net_ifs   : List of DVB network interface names
        ip_addrs  : List of IP addresses for the DVB interface slash subnet mask
        verbose   : Controls verbosity

    """
    if (verbose):
        util._print_header("Interface IP Address")

    is_root = os.geteuid() == 0
    if (not is_root):
        print("Set static IP addresses on dvbnet interfaces.\n")

    # Configure static IP addresses
    if (which("netplan") is None):
        _check_debian_net_interfaces_d(is_root)

    for net_if, ip_addr in zip(net_ifs, ip_addrs):
        _set_ip(net_if, ip_addr, verbose)

    # Bring up interfaces
    if (which("netplan") is not None):
        if (not is_root):
            print("Finally, apply the new netplan configuration:\n")
        util.run_or_print_root_cmd(["netplan", "apply"], logger)
    elif (os.path.exists("/etc/network/interfaces.d/")):
        # Debian approach
        if (not is_root):
            print(textwrap.fill("Finally, restart the networking service and "
                                "bring up the interfaces:") + "\n")
        util.run_or_print_root_cmd(["systemctl", "restart", "networking"],
                                   logger)
    elif (os.path.exists("/etc/sysconfig/network-scripts")):
        # CentOS/Fedora/RHEL approach
        if (not is_root):
            print(textwrap.fill("Finally, bring up the interfaces:") + "\n")

    if (which("netplan") is None):
        for net_if in net_ifs:
            util.run_or_print_root_cmd(["ifup", net_if], logger)


def check_ips(net_ifs, ip_addrs):
    """Check if IPs of one or multiple DVB network interface(s) are OK

    Args:
        net_ifs   : List of DVB network interface names
        ip_addrs  : List of IP addresses for the DVB interface slash subnet mask
        verbose   : Controls verbosity

    """
    for net_if, ip_addr in zip(net_ifs, ip_addrs):
        has_ip, ip_ok = _check_ip(net_if, ip_addr)
        if (not has_ip):
            raise ValueError("Interface {} does not have an IP address".format(
                net_if))
        elif (has_ip and not ip_ok):
            raise ValueError("Interface {} IP is not {}".format(net_if,
                                                                ip_addr))


def rm_ip(ifname):
    """Remove the static IP configuration of a given interface"""
    # Remove conf file for network interface (next time the interface could have
    # a different number)
    netplan_file = os.path.join("/etc/netplan/", "blocksat-" + ifname + ".yaml")
    net_file     = os.path.join("/etc/network/interfaces.d/", ifname + ".conf")
    ifcfg_file   = os.path.join("/etc/sysconfig/network-scripts/",
                                "ifcfg-" + ifname)

    if (which("netplan") is not None):
        cfg_file = netplan_file
    elif (os.path.exists(net_file)):
        cfg_file = net_file
    elif (os.path.exists(ifcfg_file)):
        cfg_file = ifcfg_file
    else:
        cfg_file = None

    if (cfg_file is not None and os.path.exists(cfg_file)):
        res = util.run_or_print_root_cmd(["rm", cfg_file])
        print("Removed configuration file {}".format(cfg_file))


def compute_rx_ips(sat_ip, n_ips, subnet="/29"):
    """Compute Rx IPs within the same subnet as the satellite Tx

    Args:
        sat_ip : satellite IP in CIDR notation
        n_ips  : Request number of IPs

    Returns:
        ips list of IPs

    """
    sat_ip_split = [x for x in sat_ip.split(".")]
    assert(len(sat_ip_split) == 4)
    base_ip     = ".".join(sat_ip_split[0:3])
    sat_ip_term = int(sat_ip[-1])
    base_offset = 3 # 3 reserved IPs for Tx host and modulator

    ips = list()
    for i in range(0, n_ips):
        rx_ip_term = sat_ip_term + base_offset + i
        ips.append(base_ip + "." + str(rx_ip_term) + subnet)
    return ips


